# Copyright © 2009-2015 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
Temporary files and directories
"""

import contextlib
import functools
import os
import shutil
import tempfile

file = functools.partial(tempfile.NamedTemporaryFile, prefix='didjvu.')
name = functools.partial(tempfile.mktemp, prefix='didjvu.')


# noinspection PyShadowingBuiltins
def hardlink(path, suffix='', prefix='didjvu.', dir=None):
    new_path = name(suffix=suffix, prefix=prefix, dir=dir)
    os.link(path, new_path)
    # noinspection PyUnresolvedReferences,PyProtectedMember
    return tempfile._TemporaryFileWrapper(
        open(new_path, 'r+b'),
        new_path
    )


@contextlib.contextmanager
def directory(*args, **kwargs):
    kwargs = dict(kwargs)
    kwargs.setdefault('prefix', 'didjvu.')
    tmpdir = tempfile.mkdtemp(*args, **kwargs)
    try:
        yield tmpdir
    finally:
        shutil.rmtree(tmpdir)


__all__ = [
    'directory',
    'file',
    'hardlink',
    'name',
]



# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

from typing import Optional


def main(prog: Optional[str] = None) -> None:
    from didjvu.didjvu import Main
    Main(prog=prog)


if __name__ == '__main__':
    main(prog='didjvu.__main__')



# Copyright © 2012-2023 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
XMP support
"""

import errno
import uuid

from didjvu import timestamp
from didjvu import version


def load_backend():
    try:
        # noinspection PyUnresolvedReferences
        from didjvu.xmp import gexiv2_backend
        return gexiv2_backend, None
    except ImportError as _error:  # no coverage
        error = _error

    try:
        # noinspection PyUnresolvedReferences
        from didjvu.xmp import libxmp_backend
        return libxmp_backend, error
    except ImportError:  # no coverage
        pass

    try:
        # noinspection PyUnresolvedReferences
        from didjvu.xmp import pyexiv2_backend
        return pyexiv2_backend, error
    except ImportError:  # no coverage
        pass

    return None, error


backend, import_error = load_backend()


def gen_uuid():
    """
    Generate a UUID URN, in accordance with RFC 4122.
    """
    # https://www.rfc-editor.org/rfc/rfc4122.html#section-3
    return f'urn:uuid:{uuid.uuid4()}'


class Event:
    def __init__(
            self,
            action=None,
            software_agent=None,
            parameters=None,
            instance_id=None,
            changed=None,
            when=None,
    ):
        if software_agent is None:
            software_agent = version.get_software_agent()
        self._items = [
            ('action', action),
            ('softwareAgent', software_agent),
            ('parameters', parameters),
            ('instanceID', instance_id),
            ('changed', changed),
            ('when', str(when)),
        ]

    @property
    def items(self):
        return iter(self._items)


def metadata(backend_module=backend):
    class Metadata(backend_module.MetadataBase):
        def update(self, media_type, internal_properties=None):
            internal_properties = internal_properties or tuple()
            instance_id = gen_uuid()
            document_id = gen_uuid()
            now = timestamp.now()
            original_media_type = self.get('dc.format')
            # TODO: try to guess original media type
            self['dc.format'] = media_type
            if original_media_type is not None:
                event_params = f'from {original_media_type} to {media_type}'
            else:
                event_params = f'to {media_type}'
            self['xmp.ModifyDate'] = now
            self['xmp.MetadataDate'] = now
            self['xmpMM.InstanceID'] = instance_id
            try:
                self['xmpMM.OriginalDocumentID']
            except KeyError:
                try:
                    original_document_id = self['xmpMM.DocumentID']
                except KeyError:
                    pass
                else:
                    self['xmpMM.OriginalDocumentID'] = original_document_id
            self['xmpMM.DocumentID'] = document_id
            event = Event(
                action='converted',
                parameters=event_params,
                instance_id=instance_id,
                when=now,
            )
            self.append_to_history(event)
            for key, value in internal_properties:
                self[f'didjvu.{key}'] = value

        def import_(self, image_filename):
            try:
                fp = open(f'{image_filename}.xmp', 'rt')
            except (OSError, IOError) as exception:
                if exception.errno == errno.ENOENT:
                    return
                raise
            try:
                self.read(fp)
            finally:
                fp.close()

        def write(self, fp):
            fp.write(self.serialize())

    return Metadata()


__all__ = [
    'backend',
    'import_error',
    'metadata',
]



# Copyright © 2010-2022 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

import errno
import locale
import os
import signal

from tests.tools import interim_environ, TestCase

from didjvu import ipc
from didjvu import temporary


NON_EXISTENT_COMMAND = 'didjvu-nonexistent-command'
UTF8_LOCALE_CANDIDATES = ['C.UTF-8', 'en_US.UTF-8']


class ExceptionsTestCase(TestCase):
    def test_valid(self):
        names = ['SIGINT', 'SIGABRT', 'SIGSEGV']
        for name in names:
            with self.subTest(name=name):
                signal_id = getattr(signal, name)
                exception = ipc.CalledProcessInterrupted(signal_id, 'eggs')
                self.assertEqual(
                    str(exception),
                    f"Command 'eggs' was interrupted by signal {name}"
                )

    def test_invalid_signal_id(self):
        # signal.NSIG is guaranteed not be a correct signal number
        exception = ipc.CalledProcessInterrupted(signal.NSIG, 'eggs')
        self.assertEqual(
            str(exception),
            f"Command 'eggs' was interrupted by signal {signal.NSIG}"
        )


class WaitTestCase(TestCase):
    def test_exit_code_0(self):
        child = ipc.Subprocess(['true'])
        return_code = child.wait()
        self.assertEqual(return_code, 0)

    def test_exit_code_1(self):
        child = ipc.Subprocess(['false'])
        with self.assertRaises(expected_exception=ipc.CalledProcessError) as exception_manager:
            child.wait()
        self.assertEqual(
            str(exception_manager.exception),
            "Command 'false' returned non-zero exit status 1."
        )

    def _test_signal(self, name):
        # Any long-standing process would do.
        with self.assertRaises(expected_exception=ipc.CalledProcessInterrupted) as exception_manager:
            with ipc.Subprocess(['cat'], stdin=ipc.PIPE) as child:
                os.kill(child.pid, getattr(signal, name))

        self.assertEqual(
            str(exception_manager.exception),
            f"Command 'cat' was interrupted by signal {name}"
        )

    def test_wait_signal(self):
        names = ['SIGINT', 'SIGABRT', 'SIGSEGV']
        for name in names:
            with self.subTest(name=name):
                self._test_signal(name)


class EnvironmentTestCase(TestCase):
    @classmethod
    def get_utf8_locale(cls):
        old_locale = locale.setlocale(locale.LC_ALL)
        try:
            for new_locale in UTF8_LOCALE_CANDIDATES:
                try:
                    locale.setlocale(locale.LC_ALL, new_locale)
                except locale.Error:
                    continue
                return new_locale
        finally:
            locale.setlocale(locale.LC_ALL, old_locale)

    def setUp(self):
        self.utf8_locale = self.get_utf8_locale()

    def test_call_without_env_parameter(self):
        with interim_environ(didjvu='42'):
            child = ipc.Subprocess(
                ['sh', '-c', 'printf $didjvu'],
                stdout=ipc.PIPE, stderr=ipc.PIPE,
            )
            stdout, stderr = child.communicate()
            self.assertEqual(stdout, b'42')
            self.assertEqual(stderr, b'')

    def test_call_with_empty_env_parameter(self):
        with interim_environ(didjvu='42'):
            child = ipc.Subprocess(
                ['sh', '-c', 'printf $didjvu'],
                stdout=ipc.PIPE, stderr=ipc.PIPE,
                env=dict(),
            )
            stdout, stderr = child.communicate()
            self.assertEqual(stdout, b'42')
            self.assertEqual(stderr, b'')

    def test_call_with_custom_env_parameter(self):
        with interim_environ(didjvu='42'):
            child = ipc.Subprocess(
                ['sh', '-c', 'printf $didjvu'],
                stdout=ipc.PIPE, stderr=ipc.PIPE,
                env=dict(didjvu='24'),
            )
            stdout, stderr = child.communicate()
            self.assertEqual(stdout, b'24')
            self.assertEqual(stderr, b'')

    def test_path(self):
        path = os.getenv('PATH')
        with temporary.directory() as tmpdir:
            command_name = temporary.name(dir=tmpdir)
            command_path = os.path.join(tmpdir, command_name)
            with open(command_path, 'wt') as file:
                print('#!/bin/sh', file=file)
                print('printf 42', file=file)
            os.chmod(command_path, 0o700)
            path = str.join(os.pathsep, [tmpdir, path])
            with interim_environ(PATH=path):
                child = ipc.Subprocess(
                    [command_name],
                    stdout=ipc.PIPE, stderr=ipc.PIPE,
                )
                stdout, stderr = child.communicate()
                self.assertEqual(stdout, b'42')
                self.assertEqual(stderr, b'')

    def _test_locale(self):
        child = ipc.Subprocess(
            ['locale'],
            stdout=ipc.PIPE, stderr=ipc.PIPE
        )
        stdout, stderr = child.communicate()
        stdout = stdout.splitlines()
        stderr = stderr.splitlines()
        self.assertEqual(stderr, [])
        data = dict(line.decode().split('=', 1) for line in stdout)
        has_lc_all = has_lc_ctype = has_lang = False
        for key, value in data.items():
            if key == 'LC_ALL':
                has_lc_all = True
                self.assertEqual(value, '')
            elif key == 'LC_CTYPE':
                has_lc_ctype = True
                if self.utf8_locale is None:
                    raise self.SkipTest(
                        f'UTF-8 locale missing '
                        f'({" or ".join(UTF8_LOCALE_CANDIDATES)})'
                    )
                self.assertEqual(value, self.utf8_locale)
            elif key == 'LANG':
                has_lang = True
                self.assertEqual(value, '')
            elif key == 'LANGUAGE':
                self.assertEqual(value, '')
            else:
                self.assertEqual(value, '"POSIX"')
        self.assertTrue(has_lc_all)
        self.assertTrue(has_lc_ctype)
        self.assertTrue(has_lang)

    def test_locale_lc_all(self):
        with interim_environ(LC_ALL=self.utf8_locale):
            self._test_locale()

    def test_locale_lc_ctype(self):
        with interim_environ(LC_ALL=None, LC_CTYPE=self.utf8_locale):
            self._test_locale()

    def test_locale_lang(self):
        with interim_environ(LC_ALL=None, LC_CTYPE=None, LANG=self.utf8_locale):
            self._test_locale()


class InitExceptionTestCase(TestCase):
    def test_init_exception(self):
        with self.assertRaises(expected_exception=OSError) as exception_manager:
            ipc.Subprocess([NON_EXISTENT_COMMAND])
        exception_message = f"[Errno {errno.ENOENT}] No such file or directory: {NON_EXISTENT_COMMAND!r}"
        self.assertEqual(str(exception_manager.exception), exception_message)


class ShellEscapeTestCase(TestCase):
    def test_no_escape(self):
        value = 'eggs'
        result = ipc.shell_escape([value])
        self.assertEqual(result, value)

    def test_escape(self):
        value = '$pam'
        result = ipc.shell_escape([value])
        self.assertEqual(result, "'$pam'")
        value = "s'pam"
        result = ipc.shell_escape([value])
        self.assertEqual(result, """'s'"'"'pam'""")

    def test_list(self):
        lst = ['$pam', 'eggs', "s'pam"]
        result = ipc.shell_escape(lst)
        self.assertEqual(result, """'$pam' eggs 's'"'"'pam'""")


class RequireTestCase(TestCase):
    # noinspection PyMethodMayBeStatic
    def test_ok(self):
        ipc.require('true', 'false')

    def test_fail(self):
        with self.assertRaises(expected_exception=OSError) as exception_manager:
            ipc.require(NON_EXISTENT_COMMAND)
        exception_message = f"[Errno {errno.ENOENT}] command not found: {NON_EXISTENT_COMMAND!r}"
        self.assertEqual(str(exception_manager.exception), exception_message)



# Copyright © 2012-2021 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

import importlib
import io
import logging
import os
from contextlib import contextmanager
from xml.etree import ElementTree

from tests.tools import SkipTest, TestCase

from didjvu import ipc
from didjvu import temporary
from didjvu import xmp
from didjvu.xmp import namespaces

logger = logging.getLogger(__name__)
del logging


def import_backend(name):
    mod_name = f'didjvu.xmp.{name}_backend'
    try:
        backend = importlib.import_module(mod_name)
    except ImportError as _import_error:
        import_error = _import_error

        class Backend:
            # Dummy replacement
            class MetadataBase:
                def __init__(self):
                    raise SkipTest(str(import_error))

        backend = Backend()
        backend.__name__ = mod_name

    return backend


XMP_BACKENDS = [
    import_backend(name)
    for name in [
        'gexiv2',
        'libxmp',
        'pyexiv2',
    ]
]

try:
    import libxmp
    import libxmp.consts
except ImportError as _libxmp_import_error:
    libxmp_import_error = _libxmp_import_error
    libxmp = None


class UuuidCheckMixin:
    UUID_REGEX = (
        r'\Aurn:uuid:XXXXXXXX-XXXX-4XXX-[89ab]XXX-XXXXXXXXXXXX\Z'
        .replace('X', '[0-9a-f]')
    )

    def assert_uuid_urn(self, uuid):
        # noinspection PyUnresolvedReferences
        self.assertRegex(
            text=uuid,
            expected_regex=self.UUID_REGEX,
        )


class UuidTestCase(UuuidCheckMixin, TestCase):
    def test_uuid(self):
        uuid1 = xmp.gen_uuid()
        self.assert_uuid_urn(uuid1)
        uuid2 = xmp.gen_uuid()
        self.assert_uuid_urn(uuid2)
        self.assertNotEqual(uuid1, uuid2)


class NamespacesTestCase(TestCase):
    def test_namespaces(self):
        if libxmp is None:
            raise self.SkipTest(str(libxmp_import_error))
        self.assertEqual(libxmp.consts.XMP_NS_DC, namespaces.dc)
        self.assertEqual(libxmp.consts.XMP_NS_RDF, namespaces.rdf)
        self.assertEqual(libxmp.consts.XMP_NS_TIFF, namespaces.tiff)
        self.assertEqual(libxmp.consts.XMP_NS_XMP, namespaces.xmp)
        self.assertEqual(libxmp.consts.XMP_NS_XMP_MM, namespaces.xmpmm)


class MetadataTestCase(UuuidCheckMixin, TestCase):
    @contextmanager
    def run_exiv2(self, filename, fail_ok=False):
        def read_from_subprocess():
            try:
                with ipc.Subprocess(
                        ['exiv2', '-P', 'Xkt', 'print', filename],
                        stdout=ipc.PIPE,
                        stderr=ipc.PIPE,
                ) as child:
                    for line in sorted(child.stdout):
                        yield line.decode('utf-8')
                    stderr = child.stderr.read()
                    if not fail_ok:
                        self.assertEqual(stderr.decode('utf-8'), '')
            except OSError as exception:
                raise self.SkipTest(str(exception))

        try:
            yield read_from_subprocess()
        except ipc.CalledProcessError:
            if not fail_ok:
                raise

    def assert_correct_software_agent(self, software_agent):
        self.assertRegex(
            text=software_agent,
            expected_regex=r'^didjvu [\d.]+( [(]Gamera [\d.]+[)])?',
        )

    def test_empty(self):
        for backend in XMP_BACKENDS:
            with temporary.file(mode='w+t') as xmp_file:
                exception = None
                try:
                    meta = xmp.metadata(backend_module=backend)
                    meta.write(xmp_file)
                    xmp_file.flush()
                    xmp_file.seek(0)
                except Exception as _exception:
                    exception = _exception

                with self.subTest(backend=backend.__name__, runner='exiv2'):
                    self._test_empty_exiv2(
                        xmp_file=xmp_file, exception=exception
                    )
                with self.subTest(backend=backend.__name__, runner='libxmp'):
                    self._test_empty_libxmp(
                        xmp_file=xmp_file, exception=exception
                    )

    def _test_empty_exiv2(self, xmp_file, exception=None):
        if exception is not None:
            raise exception
        with self.run_exiv2(xmp_file.name, fail_ok=True) as output:
            for line in output:
                self.assertEqual(line, '')

    def _test_empty_libxmp(self, xmp_file, exception=None):
        if exception is not None:
            raise exception
        if libxmp is None:
            raise self.SkipTest(str(libxmp_import_error))
        meta = libxmp.XMPMeta()
        meta.parse_from_str(xmp_file.read())
        xml_meta = meta.serialize_to_str(omit_all_formatting=True, omit_packet_wrapper=True)
        logger.debug(repr(xml_meta))
        xml_meta = io.StringIO(xml_meta)
        iterator = ElementTree.iterparse(xml_meta, events=('start', 'end'))

        def pop():
            return next(iterator)

        event, element = pop()
        self.assertEqual(event, 'start')
        self.assertEqual(element.tag, '{adobe:ns:meta/}xmpmeta')
        event, element = pop()
        self.assertEqual(event, 'start')
        self.assertEqual(element.tag, f'{{{namespaces.rdf}}}RDF')
        event, element = pop()
        self.assertEqual(event, 'start')
        self.assertEqual(element.tag, f'{{{namespaces.rdf}}}Description')
        self.assertEqual(element.attrib[f'{{{namespaces.rdf}}}about'], '')
        event, element = pop()
        self.assertEqual(event, 'end')
        event, element = pop()
        self.assertEqual(event, 'end')
        event, element = pop()
        self.assertEqual(event, 'end')
        try:
            event, element = pop()
        except StopIteration:
            event, element = None, None
        self.assertIsNone(event)

    def test_new(self):
        for backend in XMP_BACKENDS:
            with temporary.file(mode='w+t') as xmp_file:
                exception = None
                try:
                    meta = xmp.metadata(backend_module=backend)
                    meta.update(
                        media_type='image/x-test',
                        internal_properties=[
                            ('test_int', 42),
                            ('test_str', 'eggs'),
                            ('test_bool', True),
                        ]
                    )
                    meta.write(xmp_file)
                    xmp_file.flush()
                    xmp_file.seek(0)
                except Exception as _exception:
                    exception = _exception

                with self.subTest(backend=backend.__name__, runner='exiv2'):
                    self._test_new_exiv2(
                        xmp_file=xmp_file, exception=exception
                    )
                with self.subTest(backend=backend.__name__, runner='libxmp'):
                    self._test_new_libxmp(
                        xmp_file=xmp_file, exception=exception
                    )

    def _test_new_exiv2(self, xmp_file, exception=None):
        if exception is not None:
            raise exception

        with self.run_exiv2(xmp_file.name) as output:
            def pop():
                return tuple(next(output).rstrip('\n').split(None, 1))

            # Dublin Core:
            self.assertEqual(pop(), ('Xmp.dc.format', 'image/x-test'))
            # Internal properties:
            self.assertEqual(pop(), ('Xmp.didjvu.test_bool', 'True'))
            self.assertEqual(pop(), ('Xmp.didjvu.test_int', '42'))
            self.assertEqual(pop(), ('Xmp.didjvu.test_str', 'eggs'))
            # XMP:
            key, metadata_date = pop()
            self.assert_rfc3339_timestamp(metadata_date)
            self.assertEqual(key, 'Xmp.xmp.MetadataDate')
            key, modify_date = pop()
            self.assertEqual(key, 'Xmp.xmp.ModifyDate')
            self.assert_rfc3339_timestamp(modify_date)
            self.assertEqual(metadata_date, modify_date)
            # XMP Media Management:
            # - DocumentID:
            key, document_id = pop()
            self.assertEqual(key, 'Xmp.xmpMM.DocumentID')
            self.assert_uuid_urn(document_id)
            # - History:
            self.assertEqual(pop(), ('Xmp.xmpMM.History', 'type="Seq"'))
            # - History[1]:
            self.assertEqual(pop(), ('Xmp.xmpMM.History[1]', 'type="Struct"'))
            self.assertEqual(pop(), ('Xmp.xmpMM.History[1]/stEvt:action', 'converted'))
            key, event_instance_id = pop()
            self.assert_uuid_urn(event_instance_id)
            self.assertEqual(key, 'Xmp.xmpMM.History[1]/stEvt:instanceID')
            self.assertEqual(pop(), ('Xmp.xmpMM.History[1]/stEvt:parameters', 'to image/x-test'))
            key, software_agent = pop()
            self.assertEqual(key, 'Xmp.xmpMM.History[1]/stEvt:softwareAgent')
            self.assert_correct_software_agent(software_agent)
            key, event_date = pop()
            self.assertEqual((key, event_date), ('Xmp.xmpMM.History[1]/stEvt:when', modify_date))
            # - InstanceID:
            key, instance_id = pop()
            self.assertEqual(key, 'Xmp.xmpMM.InstanceID')
            self.assert_uuid_urn(instance_id)
            self.assertEqual(instance_id, event_instance_id)
            try:
                line = pop()
            except StopIteration:
                line = None
            self.assertIsNone(line)

    def _test_new_libxmp(self, xmp_file, exception=None):
        if exception is not None:
            raise exception
        if libxmp is None:
            raise self.SkipTest(str(libxmp_import_error))
        meta = libxmp.XMPMeta()

        def get(namespace, key):
            return meta.get_property(namespace, key)

        meta.parse_from_str(xmp_file.read())
        self.assertEqual(get(namespaces.dc, 'format'), 'image/x-test')
        modify_date = get(namespaces.xmp, 'ModifyDate')
        metadata_date = get(namespaces.xmp, 'MetadataDate')
        self.assertEqual(modify_date, metadata_date)
        document_id = get(namespaces.xmpmm, 'DocumentID')
        self.assert_uuid_urn(document_id)
        instance_id = get(namespaces.xmpmm, 'InstanceID')
        self.assert_uuid_urn(instance_id)
        self.assertEqual(get(namespaces.xmpmm, 'History[1]/stEvt:action'), 'converted')
        software_agent = get(namespaces.xmpmm, 'History[1]/stEvt:softwareAgent')
        self.assert_correct_software_agent(software_agent)
        self.assertEqual(get(namespaces.xmpmm, 'History[1]/stEvt:parameters'), 'to image/x-test')
        self.assertEqual(get(namespaces.xmpmm, 'History[1]/stEvt:instanceID'), instance_id)
        self.assertEqual(get(namespaces.xmpmm, 'History[1]/stEvt:when'), str(modify_date))
        self.assertEqual(get(namespaces.didjvu, 'test_int'), '42')
        self.assertEqual(get(namespaces.didjvu, 'test_str'), 'eggs')
        self.assertEqual(get(namespaces.didjvu, 'test_bool'), 'True')

    def test_updated(self):
        image_path = self.get_data_file('example.png')
        for backend in XMP_BACKENDS:
            with temporary.file(mode='w+t') as xmp_file:
                exception = None
                try:
                    meta = xmp.metadata(backend_module=backend)
                    meta.import_(image_path)
                    meta.update(
                        media_type='image/x-test',
                        internal_properties=[
                            ('test_int', 42),
                            ('test_str', 'eggs'),
                            ('test_bool', True),
                        ]
                    )
                    meta.write(xmp_file)
                    xmp_file.flush()
                    xmp_file.seek(0)
                except Exception as _exception:
                    exception = _exception

                with self.subTest(backend=backend.__name__, runner='exiv2'):
                    self._test_updated_exiv2(
                        xmp_file=xmp_file, exception=exception
                    )
                with self.subTest(backend=backend.__name__, runner='libxmp'):
                    self._test_updated_libxmp(
                        xmp_file=xmp_file, exception=exception
                    )

    _original_software_agent = 'scanhelper 0.6'
    _original_create_date = '2012-02-01T16:28:00+01:00'
    _original_document_id = 'urn:uuid:04fa0637-2b6e-417c-9fff-d6f0f02c08a6'
    _original_instance_id = 'urn:uuid:c3745412-65c0-4db4-880f-34fb57beddc0'

    def _test_updated_exiv2(self, xmp_file, exception=None):
        if exception is not None:
            raise exception

        with self.run_exiv2(xmp_file.name) as output:
            def pop():
                return tuple(next(output).rstrip('\n').split(None, 1))

            # Dublin Core:
            self.assertEqual(pop(), ('Xmp.dc.format', 'image/x-test'))
            # Internal properties:
            self.assertEqual(pop(), ('Xmp.didjvu.test_bool', 'True'))
            self.assertEqual(pop(), ('Xmp.didjvu.test_int', '42'))
            self.assertEqual(pop(), ('Xmp.didjvu.test_str', 'eggs'))
            # TIFF:
            self.assertEqual(pop(), ('Xmp.tiff.ImageHeight', '42'))
            self.assertEqual(pop(), ('Xmp.tiff.ImageWidth', '69'))
            # XMP:
            key, create_date = pop()
            self.assertEqual((key, create_date), ('Xmp.xmp.CreateDate', self._original_create_date))
            self.assertEqual(pop(), ('Xmp.xmp.CreatorTool', self._original_software_agent))
            key, metadata_date = pop()
            self.assertEqual(key, 'Xmp.xmp.MetadataDate')
            self.assert_rfc3339_timestamp(metadata_date)
            key, modify_date = pop()
            self.assertEqual(key, 'Xmp.xmp.ModifyDate')
            self.assert_rfc3339_timestamp(modify_date)
            self.assertEqual(metadata_date, modify_date)
            # XMP Media Management:
            # - DocumentID:
            key, document_id = pop()
            self.assertEqual(key, 'Xmp.xmpMM.DocumentID')
            self.assert_uuid_urn(document_id)
            self.assertNotEqual(document_id, self._original_document_id)
            # - History:
            self.assertEqual(pop(), ('Xmp.xmpMM.History', 'type="Seq"'))
            # - History[1]:
            self.assertEqual(pop(), ('Xmp.xmpMM.History[1]', 'type="Struct"'))
            self.assertEqual(pop(), ('Xmp.xmpMM.History[1]/stEvt:action', 'created'))
            key, original_instance_id = pop()
            self.assertEqual(key, 'Xmp.xmpMM.History[1]/stEvt:instanceID')
            self.assertEqual(original_instance_id, self._original_instance_id)
            self.assertEqual(pop(), ('Xmp.xmpMM.History[1]/stEvt:softwareAgent', self._original_software_agent))
            self.assertEqual(pop(), ('Xmp.xmpMM.History[1]/stEvt:when', create_date))
            # - History[2]:
            self.assertEqual(pop(), ('Xmp.xmpMM.History[2]', 'type="Struct"'))
            self.assertEqual(pop(), ('Xmp.xmpMM.History[2]/stEvt:action', 'converted'))
            key, event_instance_id = pop()
            self.assertEqual(key, 'Xmp.xmpMM.History[2]/stEvt:instanceID')
            self.assert_uuid_urn(event_instance_id)
            self.assertEqual(pop(), ('Xmp.xmpMM.History[2]/stEvt:parameters', 'from image/png to image/x-test'))
            key, software_agent = pop()
            self.assertEqual(key, 'Xmp.xmpMM.History[2]/stEvt:softwareAgent')
            self.assert_correct_software_agent(software_agent)
            self.assertEqual(pop(), ('Xmp.xmpMM.History[2]/stEvt:when', metadata_date))
            # - InstanceID:
            key, instance_id = pop()
            self.assertEqual(key, 'Xmp.xmpMM.InstanceID')
            self.assert_uuid_urn(instance_id)
            self.assertEqual(instance_id, event_instance_id)
            self.assertNotEqual(instance_id, original_instance_id)
            # - OriginalDocumentID:
            key, original_document_id = pop()
            self.assertEqual(key, 'Xmp.xmpMM.OriginalDocumentID')
            self.assertEqual(original_document_id, self._original_document_id)
            try:
                line = pop()
            except StopIteration:
                line = None
            self.assertIsNone(line)

    def _test_updated_libxmp(self, xmp_file, exception=None):
        if exception is not None:
            raise exception
        if libxmp is None:
            raise self.SkipTest(str(libxmp_import_error))
        meta = libxmp.XMPMeta()

        def get(namespace, key):
            return meta.get_property(namespace, key)

        meta.parse_from_str(xmp_file.read())
        self.assertEqual(get(namespaces.dc, 'format'), 'image/x-test')
        self.assertEqual(get(namespaces.tiff, 'ImageWidth'), '69')
        self.assertEqual(get(namespaces.tiff, 'ImageHeight'), '42')
        self.assertEqual(get(namespaces.xmp, 'CreatorTool'), self._original_software_agent)
        create_date = get(namespaces.xmp, 'CreateDate')
        self.assertEqual(create_date, self._original_create_date)
        modify_date = get(namespaces.xmp, 'ModifyDate')
        self.assertGreater(modify_date, create_date)
        metadata_date = get(namespaces.xmp, 'MetadataDate')
        self.assertEqual(modify_date, metadata_date)
        # (Original) DocumentID:
        original_document_id = get(namespaces.xmpmm, 'OriginalDocumentID')
        self.assertEqual(original_document_id, self._original_document_id)
        document_id = get(namespaces.xmpmm, 'DocumentID')
        self.assert_uuid_urn(document_id)
        self.assertNotEqual(document_id, original_document_id)
        self.assert_uuid_urn(document_id)
        # InstanceID:
        instance_id = get(namespaces.xmpmm, 'InstanceID')
        self.assert_uuid_urn(instance_id)
        # History[1]:
        self.assertEqual(get(namespaces.xmpmm, 'History[1]/stEvt:action'), 'created')
        self.assertEqual(get(namespaces.xmpmm, 'History[1]/stEvt:softwareAgent'), self._original_software_agent)
        original_instance_id = get(namespaces.xmpmm, 'History[1]/stEvt:instanceID')
        self.assertEqual(original_instance_id, self._original_instance_id)
        self.assertNotEqual(instance_id, original_instance_id)
        self.assertEqual(get(namespaces.xmpmm, 'History[1]/stEvt:when'), create_date)
        # History[2]:
        self.assertEqual(get(namespaces.xmpmm, 'History[2]/stEvt:action'), 'converted')
        software_agent = get(namespaces.xmpmm, 'History[2]/stEvt:softwareAgent')
        self.assert_correct_software_agent(software_agent)
        self.assertEqual(get(namespaces.xmpmm, 'History[2]/stEvt:parameters'), 'from image/png to image/x-test')
        self.assertEqual(get(namespaces.xmpmm, 'History[2]/stEvt:instanceID'), instance_id)
        self.assertEqual(get(namespaces.xmpmm, 'History[2]/stEvt:when'), modify_date)
        # Internal properties:
        self.assertEqual(get(namespaces.didjvu, 'test_int'), '42')
        self.assertEqual(get(namespaces.didjvu, 'test_str'), 'eggs')
        self.assertEqual(get(namespaces.didjvu, 'test_bool'), 'True')

    def _test_io_error(self, backend):
        image_path = self.get_data_file('nonexistent.png')
        meta = xmp.metadata(backend_module=backend)
        meta.import_(image_path)
        with temporary.directory() as tmpdir:
            os.chmod(tmpdir, 0o000)
            try:
                image_path = os.path.join(tmpdir, 'example.png')
                meta = xmp.metadata(backend_module=backend)
                with self.assertRaises(expected_exception=IOError):
                    meta.import_(image_path)
            finally:
                os.chmod(tmpdir, 0o700)

    def test_io_error(self):
        for backend in XMP_BACKENDS:
            with self.subTest(backend=backend):
                self._test_io_error(backend)

    def test_versions(self):
        for backend in XMP_BACKENDS:
            with self.subTest(backend=backend):
                _ = backend.MetadataBase()  # Skip if dummy.
                versions = backend.versions
                self.assertTrue(versions)
                self.assertIsInstance(versions, list)



# Copyright © 2012-2024 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
Timestamps.
"""

import datetime
import time


class Timestamp:
    def __init__(self, unix_time):
        self._localtime = time.localtime(unix_time)
        datetime_current_timezone = datetime.datetime.fromtimestamp(unix_time).astimezone()
        self._timezone_delta = datetime_current_timezone.utcoffset()

    def _str(self):
        return time.strftime('%Y-%m-%dT%H:%M:%S', self._localtime)

    def _str_timezone(self):
        offset = self._timezone_delta.days * 3600 * 24 + self._timezone_delta.seconds
        if offset == 0:
            # Apparently, pyexiv2 automatically converts 00:00 offsets to “Z”.
            # Let's always use “Z” for consistency.
            return 'Z'
        hours, minutes = divmod(abs(offset) // 60, 60)
        sign = '+' if offset >= 0 else '-'
        return f'{sign}{hours:02}:{minutes:02}'

    def __str__(self):
        """
        Format the timestamp object in accordance with RFC 3339.
        """
        return self._str() + self._str_timezone()

    def as_datetime(self, cls=datetime.datetime):
        timezone_delta = self._timezone_delta

        class Timezone(datetime.tzinfo):
            def utcoffset(self, dt):
                del dt
                return timezone_delta

            def dst(self, dt):
                del dt
                return datetime.timedelta(0)

            def tzname(self, dt):
                del dt
                return

        return cls(*self._localtime[:6], tzinfo=Timezone())


def now():
    return Timestamp(time.time())


__all__ = [
    'Timestamp',
    'now'
]



# Copyright © 2008-2022 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
Interprocess communication
"""

import errno
import logging
import os
import shlex
import re
import signal
import subprocess


# CalledProcessError, CalledProcessInterrupted
# ============================================

# Protect from scanadf[0] and possibly other software that sets
# SIGCHLD to SIG_IGN.
# [0] https://bugs.debian.org/596232
if os.name == 'posix':
    signal.signal(signal.SIGCHLD, signal.SIG_DFL)


def get_signal_names():
    signal_name_pattern = re.compile(r'^SIG[A-Z\d]*$')
    data = {
        name: getattr(signal, name)
        for name in dir(signal)
        if signal_name_pattern.match(name)
    }
    try:
        if data['SIGABRT'] == data['SIGIOT']:
            del data['SIGIOT']
    except KeyError:  # no coverage
        pass
    try:
        if data['SIGCHLD'] == data['SIGCLD']:
            del data['SIGCLD']
    except KeyError:  # no coverage
        pass
    return {number: name for name, number in data.items()}


CalledProcessError = subprocess.CalledProcessError


class CalledProcessInterrupted(CalledProcessError):
    _signal_names = get_signal_names()

    def __init__(self, signal_id, command):
        Exception.__init__(self, command, signal_id)

    def __str__(self):
        signal_name = self._signal_names.get(self.args[1], self.args[1])
        return f'Command {self.args[0]!r} was interrupted by signal {signal_name}'


del get_signal_names


# Subprocess
# ==========

def shell_escape(commandline):
    return ' '.join(map(shlex.quote, commandline))


class Subprocess(subprocess.Popen):
    @classmethod
    def override_env(cls, override):
        env = os.environ
        # We'd like to:
        # - preserve LC_CTYPE (which is required by some DjVuLibre tools),
        # - but reset all other locale settings (which tend to break things).
        lc_ctype = env.get('LC_ALL') or env.get('LC_CTYPE') or env.get('LANG')
        env = {
            key: value
            for key, value in env.items()
            if not (key.startswith('LC_') or key in {'LANG', 'LANGUAGE'})
        }
        if lc_ctype:
            env['LC_CTYPE'] = lc_ctype
        if override:
            env.update(override)
        return env

    def __init__(self, *args, **kwargs):
        kwargs['env'] = self.override_env(kwargs.get('env'))
        if os.name == 'posix':
            kwargs.update(close_fds=True)
        try:
            commandline = kwargs['args']
        except KeyError:
            commandline = args[0]
        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug(shell_escape(commandline))
        self.__command = commandline[0]
        try:
            # noinspection PyArgumentList
            subprocess.Popen.__init__(self, *args, **kwargs)
        except EnvironmentError as exception:
            exception.filename = self.__command
            raise

    def wait(self, timeout=None):
        return_code = subprocess.Popen.wait(self, timeout=timeout)
        if return_code > 0:
            raise CalledProcessError(return_code, self.__command)
        if return_code < 0:
            raise CalledProcessInterrupted(-return_code, self.__command)
        return return_code


# PIPE
# ====

PIPE = subprocess.PIPE


# require()
# =========

def _require(command):
    directories = os.environ['PATH'].split(os.pathsep)
    for directory in directories:
        path = os.path.join(directory, command)
        if os.access(path, os.X_OK):
            return
    raise OSError(errno.ENOENT, 'command not found', command)


def require(*commands):
    for command in commands:
        _require(command)


# logging support
# ===============

LOGGER = logging.getLogger('didjvu.ipc')


# __all__
# =======

__all__ = [
    'CalledProcessError', 'CalledProcessInterrupted',
    'Subprocess', 'PIPE',
]



# Copyright © 2011-2019 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
Various helper functions
"""

import os

IS_DEBIAN = os.path.exists('/etc/debian_version')


def enhance_import_error(exception, package, debian_package, homepage):
    has_debian_package = bool(IS_DEBIAN and debian_package)
    message = str(exception)
    if has_debian_package:
        package = debian_package
    message += f'; please install the {package} package'
    if not has_debian_package:
        message += f' <{homepage}>'
    exception.msg = message


class Namespace:
    pass


class Proxy:
    def __init__(self, obj, wait_fn, temporaries):
        self._object = obj
        self._wait_fn = wait_fn
        self._temporaries = temporaries

    def __getattribute__(self, name):
        if name.startswith('_'):
            return object.__getattribute__(self, name)
        self._wait_fn()
        self._wait_fn = int
        return getattr(self._object, name)

    def __setattr__(self, name, value):
        if name.startswith('_'):
            return object.__setattr__(self, name, value)
        self._wait_fn()
        self._wait_fn = int
        return setattr(self._object, name, value)


__all__ = [
    'Namespace',
    'Proxy',
    'enhance_import_error',
]



# Copyright © 2011-2024 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
didjvu version information
"""

import argparse
import sys

__version__ = '0.11.0'


def get_software_agent():
    import gamera
    result = f'didjvu {__version__}'
    result += f' (Gamera {gamera.__version__})'
    return result


class VersionAction(argparse.Action):
    """
    argparse --version action
    """

    def __init__(self, option_strings, dest=argparse.SUPPRESS):
        super(VersionAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            help="show program's version information and exit"
        )

    def __call__(self, parser, namespace, values, option_string=None):
        print(f'{parser.prog} {__version__}')
        python_version = sys.version_info
        print(f'+ Python {python_version.major}.{python_version.minor}.{python_version.micro}')

        from didjvu import gamera_support
        print(f'+ Gamera {gamera_support.gamera.__version__}')
        pil_name = 'Pillow'
        try:
            # noinspection PyUnresolvedReferences
            pil_version = gamera_support.PILImage.PILLOW_VERSION
        except AttributeError:
            try:
                # noinspection PyUnresolvedReferences
                pil_version = gamera_support.PILImage.__version__
            except AttributeError:
                pil_name = 'PIL'
                # noinspection PyUnresolvedReferences
                pil_version = gamera_support.PILImage.VERSION
        print(f'+ {pil_name} {pil_version}')

        from didjvu import xmp
        if xmp.backend:
            for version in xmp.backend.versions:
                prefix = '+'
                if version[0] == '+':
                    prefix = ' '
                print(prefix, version)

        parser.exit()


__all__ = [
    'VersionAction',
    '__version__',
    'get_software_agent',
]



# Copyright © 2012-2024 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
XMP support (GExiv2 backend)
"""

import re

# noinspection PyUnresolvedReferences
import gi
try:
    gi.require_version('GExiv2', '0.10')
    gi.require_version('GioUnix', '2.0')
except ValueError as exception:  # no coverage
    raise ImportError(exception)
# noinspection PyUnresolvedReferences
from gi.repository import GExiv2
if GExiv2.get_version() < 1202:
    # Version format: https://valadoc.org/gexiv2/GExiv2.get_version.html
    raise ImportError('GExiv2 >= 0.12.2 is required')  # no coverage

from didjvu import temporary
from didjvu import timestamp

from didjvu.xmp import namespaces


try:
    GExiv2.Metadata.try_register_xmp_namespace(namespaces.didjvu, 'didjvu')
except AttributeError:  # no coverage
    # GEXiv2 < 0.14.0
    # Might be dropped in April 2025 when Ubuntu 20.04 is EOL.
    GExiv2.Metadata.register_xmp_namespace(namespaces.didjvu, 'didjvu')


def _get_versions():
    yield f'PyGI {gi.__version__}'
    version_int = GExiv2.get_version()
    major_minor, patch = divmod(version_int, 100)
    major, minor = divmod(major_minor, 100)
    yield f'GExiv2 {major}.{minor}.{patch}'


versions = list(_get_versions())


class XmpError(RuntimeError):
    pass


class MetadataBase:
    _empty_xmp = (
        f'<x:xmpmeta xmlns:x="adobe:ns:meta/" xmlns:rdf="{namespaces.rdf}">'
        f'<rdf:RDF/>'
        f'</x:xmpmeta>'
    )

    def _read_data(self, data):
        fp = temporary.file(suffix='.xmp')
        try:
            fp.write(data.encode('utf-8'))
            fp.flush()
            self._meta = GExiv2.Metadata()
            self._meta.open_path(fp.name)
        finally:
            fp.close()

    def __init__(self):
        self._read_data(self._empty_xmp)

    def get(self, key, fallback=None):
        try:
            return self[key]
        except LookupError:
            return fallback

    def __getitem__(self, key):
        value = self._meta.try_get_tag_string(f'Xmp.{key}')
        if value is None:
            raise KeyError(f'Xmp.{key}')
        return value

    def __setitem__(self, key, value):
        if isinstance(value, timestamp.Timestamp):
            value = str(value)
        elif key.startswith('didjvu.'):
            value = str(value)
        self._meta.try_set_tag_string(f'Xmp.{key}', value)

    def _add_to_history(self, event, index):
        for key, value in event.items:
            if value is None:
                continue
            self[f'xmpMM.History[{index}]/stEvt:{key}'] = value

    def append_to_history(self, event):
        regexp = re.compile(r'^Xmp[.]xmpMM[.]History\[(\d+)\]/')
        n = 0
        for key in self._meta.get_xmp_tags():
            match = regexp.match(key)
            if match is None:
                continue
            i = int(match.group(1))
            n = max(i, n)
        if n == 0:
            self._meta.try_set_xmp_tag_struct('Xmp.xmpMM.History', GExiv2.StructureType.SEQ)
        return self._add_to_history(event, n + 1)

    def serialize(self):
        return '<?xml version="1.0"?>\n' + (
            self._meta.try_generate_xmp_packet(
                GExiv2.XmpFormatFlags.OMIT_PACKET_WRAPPER,
                0
            ) or self._empty_xmp
        )

    def read(self, fp):
        data = fp.read()
        self._read_data(data)


__all__ = [
    'MetadataBase',
    'versions',
]



# Copyright © 2010-2019 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

import gc
import sys

from tests.tools import mock, TestCase

from didjvu import utils


class EnhanceImportTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        # noinspection PyTypeChecker
        sys.modules['nonexistent'] = None

    def test_debian(self):
        with mock.patch.object(utils, 'IS_DEBIAN', True):
            with self.assertRaises(expected_exception=ImportError) as exception_manager:
                try:
                    # noinspection PyUnresolvedReferences
                    import nonexistent
                except ImportError as exception:
                    utils.enhance_import_error(
                        exception,
                        'PyNonexistent',
                        'python-nonexistent',
                        'https://pynonexistent.example.net/'
                    )
                    raise
                nonexistent.f()  # quieten pyflakes
            self.assertEqual(
                str(exception_manager.exception),
                (
                    'import of nonexistent halted; None in sys.modules; '
                    'please install the python-nonexistent package'
                )
            )

    def test_debian_without_debian_package_name(self):
        with mock.patch.object(utils, 'IS_DEBIAN', True):
            with self.assertRaises(expected_exception=ImportError) as exception_manager:
                try:
                    # noinspection PyUnresolvedReferences
                    import nonexistent
                except ImportError as exception:
                    utils.enhance_import_error(
                        exception,
                        'PyNonexistent',
                        '',
                        'https://pynonexistent.example.net/'
                    )
                    raise
                nonexistent.f()  # quieten pyflakes
            self.assertEqual(
                str(exception_manager.exception),
                (
                    'import of nonexistent halted; None in sys.modules; '
                    'please install the PyNonexistent package <https://pynonexistent.example.net/>'
                )
            )

    def test_non_debian(self):
        with mock.patch.object(utils, 'IS_DEBIAN', False):
            with self.assertRaises(expected_exception=ImportError) as exception_manager:
                try:
                    # noinspection PyUnresolvedReferences
                    import nonexistent
                except ImportError as exception:
                    utils.enhance_import_error(
                        exception,
                        'PyNonexistent',
                        'python-nonexistent',
                        'https://pynonexistent.example.net/'
                    )
                    raise
                nonexistent.f()  # quieten pyflakes
            self.assertEqual(
                str(exception_manager.exception),
                (
                    'import of nonexistent halted; None in sys.modules; '
                    'please install the PyNonexistent package <https://pynonexistent.example.net/>'
                )
            )


class ProxyTestCase(TestCase):
    def test_proxy(self):
        class Object:
            x = 42

        def wait():
            self.assertTrue(wait.ok)
            wait.ok = False

        wait.ok = False

        class Del:
            ok = False

            def __del__(self):
                type(self).ok = False

        proxy = utils.Proxy(Object, wait, [Del()])
        wait.ok = True
        self.assertEqual(proxy.x, 42)
        self.assertFalse(wait.ok)
        proxy.x = 37
        self.assertEqual(proxy.x, 37)
        self.assertEqual(Object.x, 37)
        Del.ok = True
        del proxy
        gc.collect()
        self.assertFalse(Del.ok)



# Copyright © 2012-2024 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
XMP support (pyexiv2 backend)
"""

import datetime
import itertools
import sys
from xml.etree import ElementTree

import pyexiv2.xmp

from didjvu import temporary
from didjvu import timestamp

from didjvu.xmp import namespaces


def _get_versions():
    version = pyexiv2.__version__
    yield f'pyexiv2 {version}'
    version = pyexiv2.__exiv2_version__
    yield f'+ Exiv2 {version}'


versions = list(_get_versions())


def xmp_register_namespace(prefix, uri):
    # Work-around for <https://bugs.debian.org/662878>
    # ("pyexiv2.xmp.register_namespace() requires URI ending with /")

    class Namespace(str):
        def endswith(self, suffix, *args, **kwargs):
            return True
    try:
        pyexiv2.xmp.register_namespace(Namespace(uri), prefix)
    except KeyError:
        if 'gi.repository.GExiv2' in sys.modules:
            # Most likely the namespace was registered by the GExiv2 backend.
            pass
        else:  # no coverage
            raise


xmp_register_namespace('didjvu', namespaces.didjvu)
ElementTree.register_namespace('x', 'adobe:ns:meta/')


class XmpError(RuntimeError):
    pass


def _get_almost_zero_value():
    value = 1.0
    while value / 2 > 0:
        value /= 2
    return value


class DatetimeForPyexiv2(datetime.datetime):
    __almost_zero = _get_almost_zero_value()

    def __init__(
            self, year, month, day, hour, minute, second, microsecond=0, tzinfo=None
    ):
        datetime.datetime.__init__(self)
        self.__second = second

    @property
    def second(self):
        # pyexiv2 uses HH:MM format (instead of HH:MM:SS) if .seconds is 0.
        # Let's fool it into thinking it's always non-zero.
        return self.__second or self.__almost_zero


def namespace_tag(namespace, tag):
    return f'{{{namespace}}}{tag}'


class MetadataBase:
    def _reload(self):
        fp = self._fp
        fp.flush()
        fp.seek(0)
        self._meta = pyexiv2.ImageMetadata(fp.name)
        self._meta.read()

    def _add_history(self):
        try:
            self['xmpMM.History']
        except LookupError:
            pass
        else:
            return
        self._meta.write()
        fp = self._fp
        fp.seek(0)
        xmp = ElementTree.parse(fp)
        description = None
        try:
            xmp_find = xmp.iterfind
        except AttributeError:  # no coverage
            # Python 2.6
            xmp_find = xmp.findall
        for description in xmp_find('.//' + namespace_tag(namespaces.rdf, 'Description')):
            pass
        if description is None:
            raise XmpError('Cannot add xmpMM:History')  # no coverage
        description_element = ElementTree.SubElement(description, namespace_tag(namespaces.xmpmm, 'History'))
        ElementTree.SubElement(description_element, namespace_tag(namespaces.rdf, 'Seq'))
        fp.seek(0)
        fp.truncate()
        xmp.write(fp, encoding='UTF-8')
        fp.flush()
        fp.seek(0)
        self._reload()
        try:
            self['xmpMM.History']
        except LookupError:  # no coverage
            raise XmpError('Cannot add xmpMM:History')

    def __init__(self):
        self._fp = fp = temporary.file(suffix='.xmp')
        fp.write(
            f'<x:xmpmeta xmlns:x="adobe:ns:meta/" xmlns:rdf="{namespaces.rdf}">'
            '<rdf:RDF/>'
            '</x:xmpmeta>'.encode('utf-8')
        )
        self._reload()

    def __del__(self):
        try:
            fp = self._fp
        except AttributeError:  # no coverage
            pass
        else:
            fp.close()

    def get(self, key, fallback=None):
        try:
            return self[key]
        except LookupError:
            return fallback

    def __getitem__(self, key):
        tag = self._meta[f'Xmp.{key}']
        if tag.type == 'MIMEType':
            value = tag.raw_value
        else:
            value = tag.value
        return value

    def __setitem__(self, key, value):
        if isinstance(value, timestamp.Timestamp):
            value = value.as_datetime(cls=DatetimeForPyexiv2)
        elif key.startswith('didjvu.'):
            value = str(value)
        # elif key == 'dc.format' and isinstance(value, str):
            # value = tuple(value.split('/', 1))
        self._meta[f'Xmp.{key}'] = value

    def _add_to_history(self, event, index):
        for key, value in event.items:
            if value is None:
                continue
            self[f'xmpMM.History[{index}]/stEvt:{key}'] = value

    def append_to_history(self, event):
        self._add_history()
        keys = self._meta.xmp_keys
        i = -1
        for i in itertools.count(1):
            key = f'Xmp.xmpMM.History[{i}]'
            if key not in keys:
                break
        assert i >= 0
        return self._add_to_history(event, i)

    def serialize(self):
        self._meta.write()
        fp = self._fp
        fp.seek(0)
        return fp.read().decode('utf-8')

    def read(self, file):
        data = file.read()
        fp = self._fp
        fp.seek(0)
        fp.truncate()
        fp.write(data.encode('utf-8'))
        self._reload()


__all__ = [
    'MetadataBase',
    'versions',
]



# Copyright © 2010-2015 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

from tests.tools import TestCase

from didjvu import templates


class TemplatesTestCase(TestCase):
    def test_name(self):
        path = '/path/to/eggs.png'
        memo = {}
        result = templates.expand('{name}', path, 0, memo)
        self.assertEqual(result, '/path/to/eggs.png')
        result = templates.expand('{base}', path, 0, memo)
        self.assertEqual(result, 'eggs.png')
        result = templates.expand('{name-ext}.djvu', path, 0, memo)
        self.assertEqual(result, '/path/to/eggs.djvu')
        result = templates.expand('{base-ext}.djvu', path, 0, memo)
        self.assertEqual(result, 'eggs.djvu')

    def test_page(self):
        path = '/path/to/eggs.png'
        memo = {}
        result = templates.expand('{page}', path, 0, memo)
        self.assertEqual(result, '1')
        result = templates.expand('{page:04}', path, 0, memo)
        self.assertEqual(result, '0001')
        result = templates.expand('{page}', path, 42, memo)
        self.assertEqual(result, '43')
        result = templates.expand('{page+26}', path, 42, memo)
        self.assertEqual(result, '69')
        result = templates.expand('{page-26}', path, 42, memo)
        self.assertEqual(result, '17')

    def test_bad_offset(self):
        path = '/path/to/eggs.png'
        with self.assertRaises(expected_exception=KeyError) as exception_manager:
            templates.expand('{page+ham}', path, 42, {})
        self.assertEqual(exception_manager.exception.args, ('page+ham',))

    def test_bad_type_offset(self):
        path = '/path/to/eggs.png'
        with self.assertRaises(expected_exception=KeyError) as exception_manager:
            templates.expand('{base-37}', path, 42, {})
        self.assertEqual(exception_manager.exception.args, ('base-37',))

    def test_bad_var_offset(self):
        path = '/path/to/eggs.png'
        with self.assertRaises(expected_exception=KeyError) as exception_manager:
            templates.expand('{eggs-37}', path, 42, {})
        self.assertEqual(exception_manager.exception.args, ('eggs-37',))

    def test_multi_offset(self):
        path = '/path/to/eggs.png'
        with self.assertRaises(expected_exception=KeyError) as exception_manager:
            templates.expand('{eggs+bacon+ham}', path, 42, {})
        self.assertEqual(exception_manager.exception.args, ('eggs+bacon+ham',))
        with self.assertRaises(expected_exception=KeyError) as exception_manager:
            templates.expand('{eggs-bacon-ham}', path, 42, {})
        self.assertEqual(exception_manager.exception.args, ('eggs-bacon-ham',))

    def test_duplicates(self):
        path = '/path/to/eggs.png'
        memo = {}
        result = templates.expand('{base-ext}.djvu', path, 0, memo)
        self.assertEqual(result, 'eggs.djvu')
        result = templates.expand('{base-ext}.djvu', path, 0, memo)
        self.assertEqual(result, 'eggs.1.djvu')
        result = templates.expand('{base-ext}.djvu', path, 0, memo)
        self.assertEqual(result, 'eggs.2.djvu')
        result = templates.expand('{base-ext}.2.djvu', path, 0, memo)
        self.assertEqual(result, 'eggs.2.1.djvu')
        result = templates.expand('{base-ext}.2.djvu', path, 0, memo)
        self.assertEqual(result, 'eggs.2.2.djvu')



# Copyright © 2012-2018 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
Constants for various XML namespaces
"""

didjvu = 'http://jwilk.net/software/didjvu#'

dc = 'http://purl.org/dc/elements/1.1/'
rdf = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
tiff = 'http://ns.adobe.com/tiff/1.0/'
xmp = 'http://ns.adobe.com/xap/1.0/'
xmpmm = 'http://ns.adobe.com/xap/1.0/mm/'



# Copyright © 2023-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

from tempfile import NamedTemporaryFile

from tests.tools import mock, TestCase

from didjvu.didjvu import LOGGER, Main


class MainSeparateTestCase(TestCase):
    def test_generates_output_file(self):
        source_file = self.get_data_file('greyscale-packbits.tiff')
        with NamedTemporaryFile(suffix='.djvu') as target_file, \
                mock.patch.object(LOGGER, 'info'):
            with mock.patch(
                    'sys.argv',
                    ['didjvu', 'separate', source_file, '-o', target_file.name]
            ):
                Main()
            target_file.seek(0)
            first_bytes = target_file.read(280)
            self.assertEqual(280, len(first_bytes), first_bytes)


class MainEncodeTestCase(TestCase):
    def test_generates_output_file(self):
        source_file = self.get_data_file('greyscale-packbits.tiff')
        with NamedTemporaryFile(suffix='.djvu') as target_file, \
                mock.patch.object(LOGGER, 'info'):
            with mock.patch(
                    'sys.argv',
                    ['didjvu', 'encode', source_file, '-o', target_file.name]
            ):
                Main()
            target_file.seek(0)
            first_bytes = target_file.read(340)
            self.assertEqual(340, len(first_bytes), first_bytes)



# Copyright © 2010-2018 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
Filetype detection
"""


class Generic:
    def __new__(cls, *args, **kwargs):
        pass  # no coverage

    @classmethod
    def like(cls, other):
        return issubclass(cls, other)


class Djvu(Generic):
    pass


class DjvuSingle(Djvu):
    pass


def check(filename):
    cls = Generic
    with open(filename, 'rb') as file:
        header = file.read(16)
        if header.startswith(b'AT&TFORM'):
            cls = Djvu
            if header.endswith(b'DJVU'):
                cls = DjvuSingle
    return cls


__all__ = [
    'check',
    'Djvu',
    'DjvuSingle',
]



# Copyright © 2009-2021 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
didjvu's private modules
"""



# Copyright © 2015-2024 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

import glob
import os
import re

from tests.tools import silence_truncated_file_read_warnings, TestCase

from PIL import Image

from didjvu import gamera_support


class LoadImageTestCase(TestCase):
    def _test_load_image(self, filename):
        dpi_match = re.search(r'dpi(\d+)', filename)
        path = self.get_data_file(filename)
        gamera_support.init()
        image = gamera_support.load_image(path)
        self.assertIsInstance(image, gamera_support.Image)
        if dpi_match is None:
            self.assertIsNone(image.dpi)
        else:
            dpi = int(dpi_match.group(1))
            self.assertIsInstance(image.dpi, int)
            self.assertEqual(image.dpi, dpi)

    def test_load_image(self):
        paths = []
        for extension in ['tiff', 'png', 'pgm', 'bmp']:
            paths += list(map(
                os.path.basename,
                glob.glob(self.get_data_file(f'*.{extension}'))
            ))
        for path in paths:
            with self.subTest(basename=path):
                if path == 'ycbcr-jpeg.tiff':
                    with silence_truncated_file_read_warnings():
                        self._test_load_image(path)
                else:
                    self._test_load_image(path)


class MethodsTestCase(TestCase):
    def _test_one_method(self, filename, method, kwargs):
        method = gamera_support.METHODS[method]
        path = self.get_data_file(filename)
        gamera_support.init()
        in_image = gamera_support.load_image(path)
        bin_image = method(in_image, **kwargs)
        self.assertIsInstance(bin_image, gamera_support.Image)
        self.assertEqual(bin_image.data.pixel_type, gamera_support.ONEBIT)
        self.assertEqual(in_image.dim, bin_image.dim)

    def _test_methods(self, filename):
        for method in gamera_support.METHODS:
            with self.subTest(method=method):
                kwargs = dict()
                if method == 'global':
                    kwargs = dict(threshold=42)
                self._test_one_method(filename=filename, method=method, kwargs=kwargs)

    def test_color(self):
        with silence_truncated_file_read_warnings():
            self._test_methods('ycbcr-jpeg.tiff')

    def test_grey(self):
        self._test_methods('greyscale-packbits.tiff')


class ToPilRgbTestCase(TestCase):
    def _test(self, filename):
        path = self.get_data_file(filename)
        self.assertTrue(os.path.exists(path))
        with Image.open(path) as in_image:
            if in_image.mode != 'RGB':
                in_image = in_image.convert('RGB')
            self.assertEqual(in_image.mode, 'RGB')
            gamera_support.init()
            gamera_image = gamera_support.load_image(path)
            out_image = gamera_support.to_pil_rgb(gamera_image)
            self.assert_images_equal(in_image, out_image)

    def test_color(self):
        with silence_truncated_file_read_warnings():
            self._test('ycbcr-jpeg.tiff')

    def test_grey(self):
        self._test('greyscale-packbits.tiff')


class ToPil1bppTestCase(TestCase):
    def _test(self, filename):
        path = self.get_data_file(filename)
        with Image.open(path) as in_image:
            if in_image.mode != '1':
                in_image = in_image.convert('1')
            self.assertEqual(in_image.mode, '1')
            gamera_support.init()
            gamera_image = gamera_support.load_image(path)
            out_image = gamera_support.to_pil_1bpp(gamera_image)
            out_image = out_image.convert('1')  # FIXME?
            self.assert_images_equal(in_image, out_image)

    def test_grey(self):
        self._test('greyscale-packbits.tiff')

    def test_mono(self):
        self._test('onebit.png')



# Copyright © 2009-2023 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
didjvu's command-line interface
"""

import argparse
import functools

from didjvu import djvu_support
from didjvu import version
from didjvu import xmp


def range_int(x, y, typename):
    class RangeInt(int):
        def __new__(cls, n):
            n = int(n)
            if not (x <= n <= y):
                raise ValueError
            return n
    return type(typename, (RangeInt,), {})


DPI_TYPE = range_int(djvu_support.DPI_MIN, djvu_support.DPI_MAX, 'dpi')
LOSS_LEVEL_TYPE = range_int(djvu_support.LOSS_LEVEL_MIN, djvu_support.LOSS_LEVEL_MAX, 'loss level')
SUBSAMPLE_TYPE = range_int(djvu_support.SUBSAMPLE_MIN, djvu_support.SUBSAMPLE_MAX, 'subsample')


def slice_type(max_slices=djvu_support.IW44_N_SLICES_MAX):
    def slices(value):
        result = []
        if ',' in value:
            previous_slice = 0
            for slice_ in value.split(','):
                slice_ = int(slice_)
                if slice_ <= previous_slice:
                    raise ValueError('non-increasing slice value')
                result += [slice_]
                previous_slice = slice_
        elif '+' in value:
            slice_ = 0
            for slice_increase in value.split('+'):
                slice_increase = int(slice_increase)
                if slice_increase <= 0:
                    raise ValueError('non-increasing slice value')
                slice_ += slice_increase
                result += [slice_]
        else:
            slice_ = int(value)
            if slice_ < 0:
                raise ValueError('invalid slice value')
            result = [slice_]
        assert len(result) > 0
        if len(result) > max_slices:
            raise ValueError('too many slices')
        return result
    return slices


def get_slice_repr(lst):
    def fold(inner_list, obj):
        return inner_list + [obj - sum(inner_list)]

    plus_list = functools.reduce(fold, lst[1:], lst[:1])
    return '+'.join(map(str, plus_list))


class Intact:
    def __init__(self, x):
        self.x = x

    def __call__(self):
        return self.x


def replace_underscores(s):
    return s.replace('_', '-')


def _get_method_parameters_help(methods):
    result = ['binarization methods and their parameters:']
    for name, method in sorted(methods.items()):
        result += ['  ' + name]
        for arg in method.args.values():
            arg_help = arg.name
            if arg.type in {int, float}:
                arg_help += '=' + 'NX'[arg.type is float]
                arg_help_paren = []
                if (arg.min is None) != (arg.max is None):
                    message = f'inconsistent limits for {name}.{arg.name}: min={arg.min}, max={arg.max}'
                    raise NotImplementedError(message)
                if arg.min is not None:
                    arg_help_paren += [f'{arg.min} .. {arg.max}']
                if arg.default is not None:
                    arg_help_paren += [f'default: {arg.default}']
                if arg_help_paren:
                    arg_help += f' ({", ".join(arg_help_paren)})'
            elif arg.type is bool:
                if arg.default is not False:
                    message = f'unexpected default value for {name}.{arg.name}: {arg.default}'
                    raise NotImplementedError(message)
            else:
                message = f'unexpected type for {name}.{arg.name}: {arg.type.__name__}'
                raise NotImplementedError(message)
            result += ['  - ' + arg_help]
    return '\n'.join(result)


class ArgumentParser(argparse.ArgumentParser):
    class Defaults:
        page_id_template = '{base-ext}.djvu'
        pages_per_dict = 1
        dpi = None
        foreground_slices = [100]
        foreground_crcb = djvu_support.CRCB.full
        foreground_subsample = 6
        background_slices = [74, 84, 90, 97]
        background_crcb = djvu_support.CRCB.normal
        background_subsample = 3

    def __init__(self, methods, default_method, prog=None):
        super(ArgumentParser, self).__init__(formatter_class=argparse.RawDescriptionHelpFormatter, prog=prog)
        self.add_argument('--version', action=version.VersionAction)
        parser_separate = self.add_subparser('separate', help='generate masks for images')
        parser_encode = self.add_subparser('encode', help='convert images to single-page DjVu documents')
        parser_bundle = self.add_subparser('bundle', help='convert images to bundled multi-page DjVu document')
        epilog = []
        default = self.Defaults
        for parser in (parser_separate, parser_encode, parser_bundle):
            epilog += [f'{parser.prog} --help']
            parser.add_argument('-o', '--output', metavar='FILE', help='output filename')

            if parser is parser_bundle:
                parser.add_argument(
                    '--page-id-template', metavar='TEMPLATE', default=default.page_id_template,
                    help=f'naming scheme for page identifiers (default: "{default.page_id_template}")'
                )
            else:
                if parser is parser_separate:
                    template = 'sep.{base-ext}.png'
                else:
                    template = default.page_id_template
                parser.add_argument(
                    '--output-template', metavar='TEMPLATE',
                    help=f'naming scheme for output files (e.g. "{template}")'
                )

            parser.add_argument('--losslevel', dest='loss_level', type=LOSS_LEVEL_TYPE, help=argparse.SUPPRESS)
            parser.add_argument(
                '--loss-level', dest='loss_level', type=LOSS_LEVEL_TYPE, metavar='N',
                help='aggressiveness of lossy compression'
            )
            parser.add_argument(
                '--lossless', dest='loss_level', action='store_const', const=djvu_support.LOSS_LEVEL_MIN,
                help='lossless compression (default)'
            )
            parser.add_argument(
                '--clean', dest='loss_level', action='store_const', const=djvu_support.LOSS_LEVEL_CLEAN,
                help='lossy compression: remove flyspecks'
            )
            parser.add_argument(
                '--lossy', dest='loss_level', action='store_const', const=djvu_support.LOSS_LEVEL_LOSSY,
                help='lossy compression: substitute patterns with small variations'
            )

            if parser is not parser_separate:
                parser.add_argument('--masks', nargs='+', metavar='MASK', help='use pre-generated masks')
                parser.add_argument('--mask', action='append', dest='masks', metavar='MASK', help='use a pre-generated mask')
                for layer_name, layer_id in (('foreground', 'fg'), ('background', 'bg')):
                    if layer_name == 'foreground':
                        def_slices = get_slice_repr(default.foreground_slices)
                        parser.add_argument(
                            '--fg-slices', dest='foreground_slices', type=slice_type(1), metavar='N',
                            help=f'number of slices for foreground (default: {def_slices})'
                        )
                    else:
                        def_slices = get_slice_repr(default.background_slices)
                        parser.add_argument(
                            '--bg-slices', dest='background_slices', type=slice_type(), metavar='N+...+N',
                            help=f'number of slices in each background chunk (default: {def_slices})'
                        )
                    default_crcb = getattr(default, f'{layer_name}_crcb')
                    parser.add_argument(
                        f'--{layer_id}-crcb', dest=f'{layer_name}_crcb', choices=list(map(str, djvu_support.CRCB.values)),
                        help=f'chrominance encoding for {layer_name} (default: {default_crcb})'
                    )
                    default_subsample = getattr(default, f'{layer_name}_subsample')
                    parser.add_argument(
                        f'--{layer_id}-subsample', dest=f'{layer_name}_subsample', type=SUBSAMPLE_TYPE, metavar='N',
                        help=f'subsample ratio for {layer_name} (default: {default_subsample})'
                    )
                parser.add_argument('--fg-bg-defaults', help=argparse.SUPPRESS, action='store_const', const=1)

            if parser is not parser_separate:
                parser.add_argument(
                    '-d', '--dpi', type=DPI_TYPE, metavar='N',
                    help=f'image resolution (default: {djvu_support.DPI_DEFAULT})'
                )
            if parser is parser_bundle:
                parser.add_argument(
                    '-p', '--pages-per-dict', type=int, metavar='N',
                    help=f'how many pages to compress in one pass (default: {default.pages_per_dict})'
                )
            parser.add_argument(
                '-m', '--method', choices=methods, metavar='METHOD', type=replace_underscores, default=default_method,
                help=f'binarization method (default: {default_method})'
            )
            parser.add_argument(
                '-x', '--param', action='append', dest='params', metavar='NAME[=VALUE]',
                help='binarization method parameter (can be given more than once)'
            )
            if parser is parser_encode or parser is parser_bundle:
                parser.add_argument('--xmp', action='store_true', help='create sidecar XMP metadata (experimental!)')
            parser.add_argument(
                '-v', '--verbose', dest='verbosity', action='append_const', const=None,
                help='more informational messages'
            )
            parser.add_argument(
                '-q', '--quiet', dest='verbosity', action='store_const', const=[],
                help='no informational messages'
            )
            parser.add_argument('input', metavar='IMAGE', nargs='+')
            parser.set_defaults(
                masks=[],
                fg_bg_defaults=None,
                loss_level=djvu_support.LOSS_LEVEL_MIN,
                pages_per_dict=default.pages_per_dict,
                dpi=default.dpi,
                foreground_slices=Intact(default.foreground_slices),
                background_slices=Intact(default.background_slices),
                foreground_crcb=Intact(default.foreground_crcb),
                background_crcb=Intact(default.background_crcb),
                foreground_subsample=Intact(default.foreground_subsample),
                background_subsample=Intact(default.background_subsample),
                verbosity=[None],
                xmp=False,
            )
            parser.epilog = _get_method_parameters_help(methods)
        self.epilog = 'more help:\n  ' + '\n  '.join(epilog)
        self.__methods = methods

    def add_subparser(self, name, **kwargs):
        try:
            # noinspection PyUnresolvedReferences
            self.__subparsers
        except AttributeError:
            # noinspection PyAttributeOutsideInit
            self.__subparsers = self.add_subparsers(parser_class=argparse.ArgumentParser)

        kwargs.setdefault('formatter_class', argparse.RawDescriptionHelpFormatter)
        parser = self.__subparsers.add_parser(name, **kwargs)
        parser.set_defaults(_action_=name)
        return parser

    def _parse_parameters(self, options):
        result = dict()
        for parameter in options.params or ():
            if '=' not in parameter:
                if parameter.isdigit() and len(options.method.args) == 1:
                    [parameter_name] = options.method.args
                    parameter_value = parameter
                else:
                    parameter_name = parameter
                    parameter_value = True
            else:
                [parameter_name, parameter_value] = parameter.split('=', 1)
            parameter_name = replace_underscores(parameter_name)
            try:
                argument = options.method.args[parameter_name]
            except KeyError:
                self.error(f'invalid parameter name {parameter_name!r}')
            try:
                # noinspection PyUnboundLocalVariable
                if (parameter_value is True) and (argument.type is not bool):
                    raise ValueError
                parameter_value = argument.type(parameter_value)
            except ValueError:
                self.error(f'invalid parameter {parameter_name} value: {parameter_value!r}')
            if (argument.min is not None) and parameter_value < argument.min:
                self.error(f'parameter {parameter_name} must be >= {argument.min}')
            if (argument.max is not None) and parameter_value > argument.max:
                self.error(f'parameter {parameter_name} must be <= {argument.max}')
            result[argument.name] = parameter_value
        for argument in options.method.args.values():
            if (not argument.has_default) and (argument.name not in result):
                self.error(f'parameter {argument.name} is not set')
        return result

    # noinspection PyMethodOverriding
    def parse_arguments(self, actions):
        options = argparse.ArgumentParser.parse_args(self)
        if not hasattr(options, 'fg_bg_defaults'):
            import sys
            self.print_usage()
            print('didjvu: error: too few arguments')
            sys.exit(2)
        if options.fg_bg_defaults is None:
            for layer in 'foreground', 'background':
                namespace = argparse.Namespace()
                setattr(options, f'{layer}_options', namespace)
                for facet in 'slices', 'crcb', 'subsample':
                    attribute_name = f'{layer}_{facet}'
                    value = getattr(options, attribute_name)
                    if isinstance(value, Intact):
                        value = value()
                    else:
                        options.fg_bg_defaults = False
                    setattr(namespace, facet, value)
                    delattr(options, attribute_name)
                if isinstance(namespace.crcb, str):
                    namespace.crcb = getattr(djvu_support.CRCB, namespace.crcb)
        if options.fg_bg_defaults is not False:
            options.fg_bg_defaults = True
        # noinspection PyTypeChecker
        options.verbosity = len(options.verbosity)
        if options.pages_per_dict <= 1:
            options.pages_per_dict = 1
        action = getattr(actions, vars(options).pop('_action_'))
        options.method = self.__methods[options.method]
        options.parameters = self._parse_parameters(options)
        try:
            if options.xmp and not xmp.backend:
                raise xmp.import_error  # pylint: disable=raising-bad-type
        except AttributeError:
            pass
        return action(options)


def dump_options(options, multi_page=False):
    method_name = options.method.name
    if options.params:
        method_name += ' '
        method_name += ' '.join(
            f'{parameter_name}={parameter_value}'
            for parameter_name, parameter_value
            in sorted(options.params.items())
        )
    yield 'method', method_name
    if multi_page:
        yield 'pages-per-dict', options.pages_per_dict
    yield 'loss-level', options.loss_level
    if options.fg_bg_defaults:
        yield 'fg-bg-defaults', True
    else:
        for layer_name in 'fg', 'bg':
            layer = getattr(options, f'{layer_name}_options')
            yield f'{layer_name}-crcb', str(layer.crcb)
            yield f'{layer_name}-slices', get_slice_repr(layer.slices)
            yield f'{layer_name}-subsample', layer.subsample


__all__ = [
    'ArgumentParser',
    'dump_options'
]



# Copyright © 2015 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

import io
import subprocess
import sys
from tempfile import NamedTemporaryFile

from tests.tools import mock, TestCase

from didjvu import fs


class CopyFileTestCase(TestCase):
    def _test_copy_file(self, data):
        input_file = io.StringIO(data)
        output_file = io.StringIO()
        length = fs.copy_file(input_file, output_file)
        self.assertEqual(output_file.tell(), length)
        output_file.seek(0)
        result = output_file.read()
        self.assertEqual(data, result)

    def test_copy_file(self):
        data = 'eggs'
        with self.subTest(data=data):
            self._test_copy_file(data)

        data = 'eggs' + 'spam' * 42
        with self.subTest(data=data):
            with mock.patch.object(fs, '_BLOCK_SIZE', 1):
                self._test_copy_file(data)

    def test_copy_file__to_stdout(self):
        with NamedTemporaryFile() as source_file:
            source_file.write(b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09')
            source_file.seek(0)
            result = subprocess.run(
                [
                    sys.executable,
                    '-c',
                    (
                        f'import sys; from didjvu.fs import copy_file; '
                        f'source_file = open({source_file.name!r}, mode="rb"); copy_file(source_file, sys.stdout); source_file.close()'
                    )
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            self.assertEqual(0, result.returncode, result)
            self.assertEqual(b'', result.stderr, result)
            self.assertEqual(b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09', result.stdout)


class ReplaceExtensionTestCase(TestCase):
    def test_replace_ext(self):
        result = fs.replace_ext('eggs', 'spam')
        self.assertEqual(result, 'eggs.spam')
        result = fs.replace_ext('eggs.', 'spam')
        self.assertEqual(result, 'eggs.spam')
        result = fs.replace_ext('eggs.ham', 'spam')
        self.assertEqual(result, 'eggs.spam')



# Copyright © 2009-2022 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
Bridge to the Gamera framework
"""

import collections
import ctypes
import math
import re
import sys

from didjvu import utils

try:
    from PIL import Image as PILImage
except ImportError as _exception:  # no coverage
    utils.enhance_import_error(
        _exception,
        'Pillow',
        'python3-pil',
        'https://pypi.org/project/Pillow/'
    )
    raise

try:
    import gamera
except ImportError as _exception:  # no coverage
    utils.enhance_import_error(
        _exception,
        'Gamera',
        '',
        'https://github.com/hsnr-gamera/gamera-4'
    )
    raise
del gamera  # quieten pyflakes


from gamera import __version__ as version  # noqa: E402
from gamera.core import load_image as _load_image  # noqa: E402
from gamera.core import init_gamera as _init  # noqa: E402
from gamera.core import Image, RGB, GREYSCALE, ONEBIT, Point, Dim, RGBPixel  # noqa: E402
from gamera.plugins.pil_io import from_pil as _from_pil  # noqa: E402
import gamera.args  # noqa: E402


def has_version(*required_version):
    return tuple(map(int, version.split('.'))) >= required_version


def load_image(filename):
    pil_image = PILImage.open(filename)
    dpi_x, dpi_y = pil_image.info.get('dpi', (0, 0))
    if dpi_x <= 1 or dpi_y <= 1:
        # Not reliable
        dpi = None
    else:
        dpi = int(round(
            math.hypot(dpi_x, dpi_y) / math.hypot(1, 1)
        ))
    try:
        if pil_image.format == 'TIFF':
            # Gamera handles only a few TIFF color modes correctly.
            # https://bugs.debian.org/784374
            gamera_modes = {'1', 'I;16', 'L', 'RGB'}
        elif pil_image.format == 'PNG':
            # Gamera doesn't handle 16-bit greyscale PNG images correctly.
            # https://groups.yahoo.com/neo/groups/gamera-devel/conversations/messages/2425
            gamera_modes = {'1', 'L', 'RGB'}
        else:
            gamera_modes = set()
        if pil_image.mode not in gamera_modes:
            raise IOError
        # Gamera supports more TIFF compression formats than PIL.
        # https://mail.python.org/pipermail/image-sig/2003-July/002354.html
        image = _load_image(filename)
    except IOError:
        # Gamera supports importing only 8-bit and RGB from PIL:
        if pil_image.mode[:2] in {'1', '1;', 'I', 'I;', 'L;'}:
            pil_image = pil_image.convert('L')
        elif pil_image.mode not in {'RGB', 'L'}:
            pil_image = pil_image.convert('RGB')
        assert pil_image.mode in {'RGB', 'L'}
        # noinspection PyArgumentList
        image = _from_pil(pil_image)
    pil_image.close()
    image.dpi = dpi
    return image


class Argument:
    _type_map = {
        gamera.args.Int: int,
        gamera.args.Real: float,
        gamera.args.Check: bool,
    }

    def __init__(self, arg):
        self.name = arg.name.replace(' ', '-').replace('_', '-')
        for gtype, ptype in self._type_map.items():
            if isinstance(arg, gtype):
                self.type = ptype
                break
        else:
            raise NotImplementedError(f'argument {self.name}: unsupported type {arg!r}')  # no coverage
        if self.type in {int, float}:
            self.min, self.max = arg.rng
            if self.min == -gamera.args.DEFAULT_MAX_ARG_NUMBER:
                self.min = None
            if self.max == gamera.args.DEFAULT_MAX_ARG_NUMBER:
                self.max = None
        else:
            self.min = self.max = None
        self.has_default = arg.has_default
        self.default = arg.default
        if isinstance(self.default, gamera.args.CNoneDefault):
            self.default = None
        elif self.has_default:
            if not isinstance(self.default, self.type):
                raise TypeError(
                    f'argument {self.name}: type({self.default!r}) should be {self.type.__name__}'
                )  # no coverage
        else:
            self.default = None


class Plugin:
    def __init__(self, plugin, name):
        self._plugin = plugin
        self._pixel_types = plugin.self_type.pixel_types
        self._method = None
        self.name = name
        self.args = collections.OrderedDict()
        for arg in plugin.args:
            if arg.name == 'storage format':
                continue
            arg = Argument(arg)
            self.args[arg.name] = arg

    def __call__(self, image, **kwargs):
        kwargs = {
            key.replace('-', '_'): value
            for key, value
            in kwargs.items()
        }
        pixel_types = self._pixel_types
        if image.data.pixel_type not in pixel_types:
            if RGB in pixel_types:
                image = image.to_rgb()
            elif GREYSCALE in pixel_types:
                image = image.to_greyscale()
            else:
                raise NotImplementedError(
                    f'method {self.name} does not support pixel type {image.pixel_type_name}'
                )  # no coverage
        assert image.data.pixel_type in pixel_types
        if self._method is None:
            self._method = self._plugin()
        return self._method(image, **kwargs)


def _load_methods():
    replace_suffix = re.compile(r'_threshold$').sub

    # noinspection PyUnresolvedReferences
    class MethodsContainer:
        from gamera.plugins.threshold import abutaleb_threshold
        from gamera.plugins.threshold import bernsen_threshold
        from gamera.plugins.threshold import djvu_threshold
        from gamera.plugins.threshold import otsu_threshold
        from gamera.plugins.threshold import threshold as global_threshold
        from gamera.plugins.threshold import tsai_moment_preserving_threshold as tsai
        # TODO: from gamera.plugins.binarization import gatos_threshold
        from gamera.plugins.binarization import brink_threshold
        from gamera.plugins.binarization import niblack_threshold
        from gamera.plugins.binarization import sauvola_threshold
        from gamera.plugins.binarization import shading_subtraction
        from gamera.plugins.binarization import white_rohrer_threshold

    method_mapping = dict()
    for name, plugin in vars(MethodsContainer).items():
        if name.startswith('_'):
            continue
        name = replace_suffix('', name)
        name = name.replace('_', '-')
        method = Plugin(plugin, name)
        method_mapping[name] = method
    return method_mapping


METHODS = _load_methods()


def to_pil_rgb(image):
    # About 20% faster than the standard .to_pil() method of Gamera 3.2.6.
    buffer = ctypes.create_string_buffer(3 * image.ncols * image.nrows)
    image.to_buffer(buffer)
    return PILImage.frombuffer('RGB', (image.ncols, image.nrows), buffer, 'raw', 'RGB', 0, 1)


def to_pil_1bpp(image):
    if image.data.pixel_type != GREYSCALE:
        image = image.to_greyscale()
    return image.to_pil()


def init():
    if not has_version(4, 0):
        raise RuntimeError('Gamera >= 4.0 is required')
    blocked_numpy = False
    if 'numpy' not in sys.modules:
        # noinspection PyTypeChecker
        sys.modules['numpy'] = None
        blocked_numpy = True
    try:
        result = _init()
    finally:
        if blocked_numpy:
            del sys.modules['numpy']
    return result


__all__ = [
    # classes:
    'Dim',
    'Image',
    'Point',
    'RGBPixel',
    # pixel types:
    'ONEBIT',
    'GREYSCALE',
    'RGB',
    # functions:
    'init',
    'load_image',
    'METHODS',
    'to_pil_1bpp',
    'to_pil_rgb',
]



# Copyright © 2010-2018 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
Filename templates
"""

import os
import string


FORMATTER = string.Formatter()


def expand(template, name, page, memo):
    base = os.path.basename(name)
    name_ext, _ = os.path.splitext(name)
    base_ext, _ = os.path.splitext(base)
    d = {
        'name': name,
        'name-ext': name_ext,
        'base': base,
        'base-ext': base_ext,
        'page': page + 1,
    }
    for _, var, _, _ in FORMATTER.parse(template):
        if var is None:
            continue
        if '+' in var:
            sign = +1
            base_var, offset = var.split('+', 1)
        elif '-' in var:
            sign = -1
            base_var, offset = var.split('-', 1)
        else:
            continue  # no coverage
            # https://bitbucket.org/ned/coveragepy/issues/198
        try:
            offset = sign * int(offset, 10)
        except ValueError:
            continue
        try:
            base_value = d[base_var]
        except LookupError:
            continue
        if not isinstance(base_value, int):
            continue
        d[var] = d[base_var] + offset
    ident = FORMATTER.vformat(template, (), d)
    while True:
        n = memo.get(ident, 0)
        if n == 0:
            break
        memo[ident] += 1
        ident_base, ident_ext = os.path.splitext(ident)
        ident = f'{ident_base}.{n}{ident_ext}'
    assert ident not in memo
    memo[ident] = 1
    return ident


__all__ = ['expand']



# Copyright © 2009-2022 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
Wrappers for the DjVuLibre utilities
"""

import os
import re
import struct

from didjvu import ipc
from didjvu import temporary
from didjvu import utils

DPI_MIN = 72
DPI_DEFAULT = 300
DPI_MAX = 6000

LOSS_LEVEL_MIN = 0
LOSS_LEVEL_CLEAN = 1
LOSS_LEVEL_LOSSY = 100
LOSS_LEVEL_MAX = 200

SUBSAMPLE_MIN = 1
SUBSAMPLE_DEFAULT = 3
SUBSAMPLE_MAX = 12

IW44_SLICES_DEFAULT = (74, 89, 99)
IW44_N_SLICES_MAX = 63
# https://sourceforge.net/p/djvu/djvulibre-git/ci/release.3.5.27.1/tree/tools/c44.cpp#l246


class Crcb:
    def __init__(self, sort_key, name):
        self._sort_key = sort_key
        self._name = name

    def __str__(self):
        return self._name


class CRCB:
    full = Crcb(3, 'full')
    normal = Crcb(2, 'normal')
    half = Crcb(1, 'half')
    none = Crcb(0, 'none')

    values = [full, normal, half, none]


def bitonal_to_djvu(image, dpi=300, loss_level=0):
    pbm_file = temporary.file(suffix='.pbm')
    image.save(pbm_file.name)
    djvu_file = temporary.file(suffix='.djvu', mode='r+b')
    args = [
        'cjb2',
        '-dpi', str(dpi),
        '-losslevel', str(loss_level),
        pbm_file.name,
        djvu_file.name
    ]

    def wait_function():
        ipc.Subprocess(args).wait()
        pbm_file.close()

    return utils.Proxy(djvu_file, wait_function, [pbm_file])


def photo_to_djvu(image, dpi=100, slices=IW44_SLICES_DEFAULT, gamma=2.2, mask_image=None, crcb=CRCB.normal):
    ppm_file = temporary.file(suffix='.ppm')
    image.save(ppm_file.name)
    if not isinstance(crcb, Crcb):
        raise TypeError
    with temporary.directory() as djvu_dir:
        args = [
            'c44',
            '-dpi', str(dpi),
            '-slice', ','.join(map(str, slices)),
            '-gamma', f'{gamma:.1f}',
            f'-crcb{crcb}',
        ]
        if mask_image is not None:
            pbm_file = temporary.file(suffix='.pbm')
            mask_image.save(pbm_file.name)
            args += ['-mask', pbm_file.name]
        djvu_path = os.path.join(djvu_dir, 'result.djvu')
        args += [ppm_file.name, djvu_path]
        ipc.Subprocess(args).wait()
        if mask_image is not None:
            pbm_file.close()
        ppm_file.close()
        return temporary.hardlink(djvu_path, suffix='.djvu')


def djvu_to_iw44(djvu_file):
    # TODO: Use Multichunk.
    iw44_file = temporary.file(suffix='.iw44')
    args = ['djvuextract', djvu_file.name, 'BG44=' + iw44_file.name]
    with open(os.devnull, 'wb') as dev_null:
        return utils.Proxy(iw44_file, ipc.Subprocess(args, stderr=dev_null).wait, [djvu_file])


def _int_or_none(x):
    if x is None:
        return
    if isinstance(x, int):
        return x
    raise TypeError


def _chunk_order(key):
    # INCL must go before Sjbz.
    if key[0] == 'incl':
        return -2
    # djvuextract expects Sjbz before PPM.
    if key[0] == 'sjbz':
        return -1
    return 0


class Multichunk:
    _chunk_names = 'Sjbz Smmr BG44 BGjp BG2k FGbz FG44 FGjp FG2k INCL Djbz'
    _chunk_names = {x.lower(): x for x in _chunk_names.split()}
    _info_re = re.compile(r' (\d+)x(\d+),.* (\d+) dpi,').search

    def __init__(self, width=None, height=None, dpi=None, **chunks):
        self.width = _int_or_none(width)
        self.height = _int_or_none(height)
        self.dpi = _int_or_none(dpi)
        self._chunks = {}
        self._dirty = set()  # Chunks that need to be re-read from the file.
        self._pristine = False  # Should save() be a no-op?
        self._file = None
        for key, value in chunks.items():
            self[key] = value

    def _load_file(self):
        args = ['djvudump', self._file.name]
        dump = ipc.Subprocess(args, stdout=ipc.PIPE)
        self.width = self.height = self.dpi = None
        keys = set()
        try:
            header = dump.stdout.readline()
            if not header.startswith('  FORM:DJVU '):
                raise ValueError
            for line in dump.stdout:
                if line[:4] == '    ' and line[8:9] == ' ':
                    key = line[4:8]
                    if key == 'INFO':
                        m = self._info_re(line[8:])
                        self.width, self.height, self.dpi = m.groups()
                    else:
                        keys.add(key.lower())
                else:
                    raise ValueError
        finally:
            dump.wait()
        self._chunks = {key: None for key in keys}
        self._dirty.add(self._chunks)
        self._pristine = True

    @classmethod
    def from_file(cls, djvu_file):
        self = cls()
        self._file = djvu_file
        self._load_file()
        return self

    def __contains__(self, key):
        key = key.lower()
        if key in self._chunks:
            return True

    def __setitem__(self, key, value):
        if key == 'image':
            ppm_file = temporary.file(suffix='.ppm')
            value.save(ppm_file.name)
            key = 'PPM'
            value = ppm_file
        else:
            key = key.lower()
            if key not in self._chunk_names:
                raise ValueError
        self._chunks[key] = value
        self._dirty.discard(key)
        self._pristine = False

    def __getitem__(self, key):
        key = key.lower()
        value = self._chunks[key]
        if key in self._dirty:
            self.save()
            self._load_file()
            self._update_chunks()
            value = self._chunks[key]
            assert value is not None
        return value

    def _update_chunks(self):
        assert self._file is not None
        args = ['djvuextract', self._file.name]
        chunk_files = {}
        for key in self._dirty:
            chunk_file = temporary.file(suffix=f'.{key}-chunk')
            args += [f'{self._chunk_names[key]}={chunk_file.name}']
            chunk_files[key] = chunk_file
        with open(os.devnull, 'wb') as dev_null:
            djvu_extract = ipc.Subprocess(args, stderr=dev_null)
        for key in self._chunks:
            self._chunks[key] = utils.Proxy(chunk_files[key], djvu_extract.wait, [self._file])
            self._dirty.discard(key)
        assert not self._dirty
        # The file reference is not needed anymore.
        self._file = None

    def save(self):
        if (self._file is not None) and self._pristine:
            return self._file
        if self.width is None:
            raise ValueError
        if self.height is None:
            raise ValueError
        if self.dpi is None:
            raise ValueError
        if len(self._chunks) == 0:
            raise ValueError
        args = ['djvumake', None, f'INFO={self.width},{self.height},{self.dpi}']
        incl_dir = None
        for key, value in sorted(self._chunks.items(), key=_chunk_order):
            try:
                key = self._chunk_names[key]
            except KeyError:
                pass
            if key == 'INCL':
                # This is tricky. DjVuLibre used to require (at least until v3.5.27) full path for INCL,
                # but now it requires just basename:
                # https://sourceforge.net/p/djvu/djvulibre-git/ci/2fea3cdd3eb2b9ae60a43a851dbd838b6939af4b/
                # Let's work around this inconsistency.
                new_incl_dir = os.path.dirname(value)
                if incl_dir is None:
                    incl_dir = new_incl_dir
                elif incl_dir != new_incl_dir:
                    raise ValueError
                value = os.path.basename(value)
            if not isinstance(value, str):
                value = value.name
            if key == 'BG44':
                value += ':99'
            args += [f'{key}={value}']

        def chdir():
            os.chdir(incl_dir or '.')

        with temporary.directory() as tmpdir:
            djvu_filename = args[1] = os.path.join(tmpdir, 'result.djvu')
            ipc.Subprocess(args, preexec_fn=chdir).wait()
            self._chunks.pop('PPM', None)
            assert 'PPM' not in self._dirty
            self._file = temporary.hardlink(djvu_filename)
            self._pristine = True
            return self._file

    def close(self):
        if 'sjbz' in self:
            self['sjbz'].close()
        if 'PPM' in self:
            self['PPM'].close()
        for chunk in self._chunks.values():
            chunk.close()


_DJVU_HEADER = b'AT&TFORM\0\0\0\0DJVMDIRM\0\0\0\0\1'


def bundle_djvu_via_indirect(*component_filenames):
    with temporary.directory() as tmpdir:
        page_ids = []
        page_sizes = []
        for filename in component_filenames:
            page_id = os.path.basename(filename)
            os.symlink(filename, os.path.join(tmpdir, page_id))
            page_ids += [page_id]
            page_size = os.path.getsize(filename)
            if page_size >= 1 << 24:
                # Would overflow; but 0 is fine, too.
                page_size = 0
            page_sizes += [page_size]
        with temporary.file(dir=tmpdir, suffix='djvu') as index_file:
            index_file.write(_DJVU_HEADER)
            index_file.write(struct.pack('>H', len(page_ids)))
            index_file.flush()
            bzz = ipc.Subprocess(['bzz', '-e', '-', '-'], stdin=ipc.PIPE, stdout=index_file)
            try:
                for page_size in page_sizes:
                    bzz.stdin.write(struct.pack('>I', page_size)[1:])
                for page_id in page_ids:
                    bzz.stdin.write(struct.pack('B', not page_id.endswith('.iff')))
                for page_id in page_ids:
                    bzz.stdin.write(page_id.encode('utf-8'))
                    bzz.stdin.write(b'\0')
            finally:
                bzz.stdin.close()
                bzz.wait()
            index_file_size = index_file.tell()
            i = 0
            while True:
                i = _DJVU_HEADER.find(b'\0' * 4, i)
                if i < 0:
                    break
                index_file.seek(i)
                index_file.write(struct.pack('>I', index_file_size - i - 4))
                i += 4
            index_file.flush()
            djvu_file = temporary.file(suffix='.djvu')
            ipc.Subprocess(['djvmcvt', '-b', index_file.name, djvu_file.name]).wait()
    return djvu_file


def bundle_djvu(*component_filenames):
    assert len(component_filenames) > 0
    if any(c.endswith('.iff') for c in component_filenames):
        # We can't use `djvm -c`.
        return bundle_djvu_via_indirect(*component_filenames)
    else:
        djvu_file = temporary.file(suffix='.djvu')
        args = ['djvm', '-c', djvu_file.name]
        args += component_filenames
        return utils.Proxy(djvu_file, ipc.Subprocess(args).wait, None)


def require_cli():
    ipc.require(
        'cjb2',
        'c44',
        'djvuextract',
        'djvudump',
        'djvumake',
        'bzz',
        'djvmcvt',
    )


_PAGE_ID_CHARACTERS = re.compile(r'^[A-Za-z\d_+.-]*$').match


def validate_page_id(page_id):
    if not _PAGE_ID_CHARACTERS(page_id):
        raise ValueError('page identifier must consist only of lowercase ASCII letters, digits, _, +, - and dot')
    if page_id[:1] in {'.', '+', '-'}:
        raise ValueError('page identifier cannot start with +, - or a dot')
    if '..' in page_id:
        raise ValueError('page identifier cannot contain two consecutive dots')
    assert page_id == os.path.basename(page_id)
    if page_id.endswith('.djvu'):
        return page_id
    else:
        raise ValueError('page identifier must end with the .djvu extension')


__all__ = [
    'bitonal_to_djvu', 'photo_to_djvu', 'djvu_to_iw44',
    'bundle_djvu',
    'require_cli',
    'validate_page_id',
    'Multichunk',
    'DPI_MIN', 'DPI_DEFAULT', 'DPI_MAX',
    'LOSS_LEVEL_MIN', 'LOSS_LEVEL_CLEAN', 'LOSS_LEVEL_LOSSY', 'LOSS_LEVEL_MAX',
    'SUBSAMPLE_MIN', 'SUBSAMPLE_DEFAULT', 'SUBSAMPLE_MAX',
    'IW44_SLICES_DEFAULT',
    'CRCB',
]



# Copyright © 2015-2021 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

import collections
import contextlib
import io
import shutil
import subprocess
import sys

from tests.tools import mock, TestCase

from didjvu import cli


class RangeIntTestCase(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.type_ = cli.range_int(37, 42, 'eggs')

    def test_less_than_minimum(self):
        with self.assertRaises(expected_exception=ValueError):
            self.type_('36')

    def test_minimum(self):
        self.assertEqual(self.type_('37'), 37)

    def test_maximum(self):
        self.assertEqual(self.type_('42'), 42)

    def test_greater_than_maximum(self):
        with self.assertRaises(expected_exception=ValueError):
            self.type_('43')

    def test_non_int(self):
        with self.assertRaises(expected_exception=ValueError):
            self.type_('')
        with self.assertRaises(expected_exception=ValueError):
            self.type_('ham')


class SliceTypeTestCase(TestCase):
    def setUp(self):
        self.type_ = cli.slice_type()
        self.type_with_max_slices_1 = cli.slice_type(max_slices=1)

    def test_non_int(self):
        with self.assertRaises(expected_exception=ValueError):
            self.type_('')
        with self.assertRaises(expected_exception=ValueError):
            self.type_('ham')

    def test_negative(self):
        with self.assertRaises(expected_exception=ValueError):
            self.type_('-42')

    def test_zero(self):
        self.assertEqual(self.type_('0'), [0])

    def test_zero_1(self):
        self.assertEqual(self.type_with_max_slices_1('0'), [0])

    def test_positive(self):
        self.assertEqual(self.type_('42'), [42])

    def test_positive_1(self):
        self.assertEqual(self.type_with_max_slices_1('42'), [42])

    def test_comma(self):
        self.assertEqual(self.type_('37,42'), [37, 42])

    def test_comma_1(self):
        with self.assertRaises(expected_exception=ValueError):
            self.type_with_max_slices_1('37,42')

    def test_comma_equal(self):
        with self.assertRaises(expected_exception=ValueError):
            self.type_('37,37')

    def test_comma_less_than(self):
        with self.assertRaises(expected_exception=ValueError):
            self.type_('42,37')

    def test_plus(self):
        self.assertEqual(self.type_('37+5'), [37, 42])

    def test_plus_1(self):
        with self.assertRaises(expected_exception=ValueError):
            self.type_with_max_slices_1('37+5')

    def test_plus_equal(self):
        with self.assertRaises(expected_exception=ValueError):
            self.type_('37+0')

    def test_plus_less_than(self):
        with self.assertRaises(expected_exception=ValueError):
            self.type_('42+-5')


class SliceRepresentationTestCase(TestCase):
    def test_representation(self):
        type_ = cli.slice_type()
        samples = [
            ([0], '0'),
            ([42], '42'),
            ([23, 37, 42], '23+14+5')
        ]

        for input_list, expected_representation in samples:
            with self.subTest(input_list=input_list):
                representation = cli.get_slice_repr(input_list)
                self.assertEqual(representation, expected_representation)
                self.assertEqual(type_(representation), input_list)


class IntactTestCase(TestCase):
    def test_intact(self):
        x = object()
        intact_x = cli.Intact(x)
        self.assertIs(intact_x(), x)


class ReplaceUnderscoresTestCase(TestCase):
    def test_replace_underscores(self):
        self.assertEqual(
            cli.replace_underscores('eggs_ham_spam'),
            'eggs-ham-spam'
        )


class MockActions(object):
    def __getattr__(self, name):
        def f(options):
            return name, options
        return f


class MockMethod(object):
    def __init__(self):
        self.args = dict()


class ArgumentParserTestCase(TestCase):
    def setUp(self):
        self.methods = dict(
            abutaleb=MockMethod(),
            djvu=MockMethod(),
        )
        self.actions = MockActions()

        self.action_names = collections.OrderedDict()
        self.action_names['separate'] = 1
        self.action_names['encode'] = 1
        self.action_names['bundle'] = 1

    def test_init(self):
        cli.ArgumentParser(self.methods, 'djvu')

    def test_no_args(self):
        result = subprocess.run(
            [shutil.which('didjvu')],
            capture_output=True,
            text=True,
        )
        self.assertEqual(2, result.returncode, result)
        actions = ','.join(self.action_names)
        self.assertMultiLineEqual(
            result.stdout,
            (
                f'usage: didjvu [-h] [--version] {{{actions}}} ...\n'
                'didjvu: error: too few arguments\n'
            )
        )

    def _test_action_no_args(self, action):
        result = subprocess.run(
            [shutil.which('didjvu'), action],
            capture_output=True,
            text=True,
        )
        self.assertEqual(2, result.returncode, result)
        self.assertRegex(
            result.stderr,
            (
                r'(?s)\A'
                f'usage: didjvu {action} .*\n'
                f'didjvu {action}: error: the following arguments are required: .*'
                r'\Z'
            )
        )

    def test_action_no_args(self):
        actions = ['separate', 'bundle', 'encode']

        for action in actions:
            with self.subTest(action=action):
                self._test_action_no_args(action=action)

    def test_bad_action(self, action='eggs'):
        result = subprocess.run(
            [shutil.which('didjvu'), action],
            capture_output=True,
            text=True,
        )
        self.assertEqual(2, result.returncode, result)

        action_values = ','.join(self.action_names)
        if sys.version_info < (3, 12, 8):
            # This unfortunately has been changed in a patch release:
            # https://github.com/python/cpython/commit/21524eec48f5b1c807f185253e9350cfdd897ce0
            action_strings = ', '.join(map(repr, self.action_names))
        else:
            action_strings = ', '.join(map(str, self.action_names))
        self.assertMultiLineEqual(
            (
                f'usage: didjvu [-h] [--version] {{{action_values}}} ...\n'
                f"didjvu: error: argument {{{action_values}}}: invalid choice: 'eggs' (choose from {action_strings})\n"
            ),
            result.stderr,
        )

    def _test_action(self, action, *args):
        stderr = io.StringIO()
        argv = ['didjvu', action]
        argv += args
        with mock.patch('sys.argv', argv), contextlib.redirect_stderr(stderr):
            parser = cli.ArgumentParser(self.methods, 'djvu')
            selected_action, options = parser.parse_arguments(self.actions)
        self.assertMultiLineEqual(stderr.getvalue(), '')
        self.assertEqual(selected_action, action)
        return options

    def _test_action_defaults(self, action):
        path = 'eggs.png'
        options = self._test_action(action, path)
        self.assertEqual(options.input, [path])
        self.assertEqual(options.masks, [])
        self.assertIsNone(options.output)
        if action == 'bundle':
            self.assertEqual(options.page_id_template, '{base-ext}.djvu')
        else:
            self.assertIsNone(options.output_template)
        self.assertIsNone(options.dpi)
        self.assertTrue(options.fg_bg_defaults)
        self.assertEqual(options.loss_level, 0)
        self.assertEqual(options.pages_per_dict, 1)
        self.assertIs(options.method, self.methods['djvu'])
        self.assertEqual(options.parameters, {})
        self.assertEqual(options.verbosity, 1)
        self.assertIs(options.xmp, False)

    def test_action_defaults(self):
        actions = ['separate', 'bundle', 'encode']

        for action in actions:
            with self.subTest(action=action):
                self._test_action_defaults(action=action)

    def _test_help(self, action=None):
        argv = ['didjvu', action, '--help']
        argv = [_f for _f in argv if _f]
        stdout = io.StringIO()
        with mock.patch('sys.argv', argv), contextlib.redirect_stdout(stdout):
            parser = cli.ArgumentParser(self.methods, 'djvu')
            with self.assertRaises(expected_exception=SystemExit) as exception_manager:
                parser.parse_arguments(dict())
            self.assertEqual(exception_manager.exception.args, (0,))
        self.assertGreater(len(stdout.getvalue()), 0)

    def test_help(self):
        actions = [None, 'separate', 'bundle', 'encode']

        for action in actions:
            with self.subTest(action=action):
                self._test_help(action=action)



# Copyright © 2012-2024 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
XMP support (python-xmp-toolkit backend)
"""

import libxmp

from didjvu import timestamp

from didjvu.xmp import namespaces


def _get_versions():
    library = 'python-xmp-toolkit'
    try:
        version = libxmp.__version__
    except AttributeError:
        return library
    return f'{library} {version}'


versions = [_get_versions()]


class XmpError(RuntimeError):
    pass


class MetadataBase:
    def __init__(self):
        backend = self._backend = libxmp.XMPMeta()
        prefix = backend.register_namespace(namespaces.didjvu, 'didjvu')
        if prefix is None:
            raise XmpError('Cannot register namespace for didjvu internal properties')  # no coverage

    @classmethod
    def _expand_key(cls, key):
        namespace, key = key.split('.', 1)
        namespace = getattr(namespaces, namespace.lower())
        return namespace, key

    def get(self, key, fallback=None):
        namespace, key = self._expand_key(key)
        backend = self._backend
        if backend.does_property_exist(namespace, key):
            result = backend.get_property(namespace, key)
        else:
            result = None
        if result is None:
            result = fallback
        return result

    def __getitem__(self, key):
        result = self.get(key)
        if result is None:
            raise KeyError(key)
        return result

    def __setitem__(self, key, value):
        namespace, key = self._expand_key(key)
        backend = self._backend
        if isinstance(value, bool):
            return_code = backend.set_property_bool(namespace, key, value)
        elif isinstance(value, int):
            return_code = backend.set_property_int(namespace, key, value)
        elif isinstance(value, list) and len(value) == 0:
            return_code = backend.set_property(
                namespace, key, '',
                prop_value_is_array=True,
                prop_array_is_ordered=True
            )
        else:
            if isinstance(value, timestamp.Timestamp):
                value = str(value)
            return_code = backend.set_property(namespace, key, value)
        if return_code is False:
            raise XmpError('Cannot set property')  # no coverage

    def _add_to_history(self, event, index):
        for key, value in event.items:
            if value is None:
                continue
            self[f'xmpMM.History[{index}]/stEvt:{key}'] = value

    def append_to_history(self, event):
        backend = self._backend

        def count_history():
            return backend.count_array_items(namespaces.xmpmm, 'History')

        count = count_history()
        if count == 0:
            self['xmpMM.History'] = []
            assert count_history() == 0

        self._add_to_history(event, count + 1)
        assert count_history() == count + 1

    def serialize(self):
        backend = self._backend
        return backend.serialize_and_format(omit_packet_wrapper=True, tabchr='    ')

    def read(self, fp):
        backend = self._backend
        xmp = fp.read()
        backend.parse_from_str(xmp)


__all__ = [
    'MetadataBase',
    'versions',
]



# Copyright © 2015 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

import os

from tests.tools import TestCase

from didjvu import filetype


class FileTypeTestCase(TestCase):
    def test_djvu(self):
        path = self.get_data_file('ycbcr.djvu')
        file_type = filetype.check(path)
        self.assertTrue(file_type.like(filetype.Djvu))
        self.assertTrue(file_type.like(filetype.DjvuSingle))

    def test_bad(self):
        path = self.get_data_file(os.devnull)
        file_type = filetype.check(path)
        self.assertFalse(file_type.like(filetype.Djvu))
        self.assertFalse(file_type.like(filetype.DjvuSingle))



# Copyright © 2009-2021 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

"""
didjvu core
"""

import itertools
import logging
import os
import sys
from typing import Any, Dict, Tuple

from didjvu import cli
from didjvu import djvu_support
from didjvu import filetype
from didjvu import fs
from didjvu import gamera_support
from didjvu import ipc
from didjvu import templates
from didjvu import temporary
from didjvu import utils
from didjvu import xmp


def setup_logging() -> Tuple[logging.Logger, logging.Logger]:
    logger = logging.getLogger('didjvu.main')
    ipc_logger = logging.getLogger('didjvu.ipc')
    logging.NOSY = (logging.INFO + logging.DEBUG) // 2  # type: ignore[attr-defined]
    # noinspection PyUnresolvedReferences
    assert logging.INFO > logging.NOSY > logging.DEBUG  # type: ignore[attr-defined]

    def nosy(msg: str, *args: Any, **kwargs: Any):
        # noinspection PyUnresolvedReferences
        logger.log(logging.NOSY, msg, *args, **kwargs)  # type: ignore[attr-defined]

    logger.nosy = nosy  # type: ignore[attr-defined]
    # Main handler:
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(message)s')
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    # IPC handler:
    handler = logging.StreamHandler()
    formatter = logging.Formatter('+ %(message)s')
    handler.setFormatter(formatter)
    ipc_logger.addHandler(handler)

    return logger, ipc_logger


LOGGER, IPC_LOGGER = setup_logging()


def error(message: str, *args: Any, **kwargs: Any) -> None:
    if args or kwargs:
        message = message.format(*args, **kwargs)
    print(f'didjvu: error: {message}', file=sys.stderr)
    sys.exit(1)


def parallel_for(options, function, *iterables):
    for args in zip(*iterables):
        function(options, *args)


def check_tty():
    if sys.stdout.isatty():
        error('refusing to write binary data to a terminal')


def get_subsampled_dimensions(image, subsample):
    width = (image.ncols + subsample - 1) // subsample
    height = (image.nrows + subsample - 1) // subsample
    return gamera_support.Dim(width, height)


def subsample_foreground(image, mask, options):
    # TODO: Optimize, perhaps using Cython.
    ratio = options.subsample
    subsampled_size = get_subsampled_dimensions(mask, ratio)
    mask = mask.to_greyscale()
    mask = mask.threshold(254)
    mask = mask.erode()
    subsampled_image = gamera_support.Image((0, 0), subsampled_size, pixel_type=gamera_support.RGB)
    subsampled_mask = gamera_support.Image((0, 0), subsampled_size, pixel_type=gamera_support.ONEBIT)
    y0 = 0
    width, height = image.ncols, image.nrows
    image = image.to_rgb()
    image_get = image.get
    mask_get = mask.get
    subsampled_image_set = subsampled_image.set
    subsampled_mask_set = subsampled_mask.set
    for sy in range(0, subsampled_image.nrows):
        x0 = 0
        for sx in range(0, subsampled_image.ncols):
            n = r = g = b = 0
            y = y0
            for dy in range(ratio):
                if y >= height:
                    break
                x = x0
                for dx in range(ratio):
                    if x >= width:
                        break
                    pt = gamera_support.Point(x, y)
                    if mask_get(pt):
                        n += 1
                        pixel = image_get(pt)
                        r += pixel.red
                        g += pixel.green
                        b += pixel.blue
                    x += 1
                y += 1
            pt = gamera_support.Point(sx, sy)
            if n > 0:
                r = (r + n // 2) // n
                g = (g + n // 2) // n
                b = (b + n // 2) // n
                subsampled_image_set(pt, gamera_support.RGBPixel(r, g, b))
            else:
                subsampled_mask_set(pt, 1)
            x0 += ratio
        y0 += ratio
    return subsampled_image, subsampled_mask


def subsample_background(image, mask, options):
    dim = get_subsampled_dimensions(mask, options.subsample)
    mask = mask.to_greyscale()
    mask = mask.resize(dim, 0)
    mask = mask.dilate().dilate()
    mask = mask.threshold(254)
    image = image.resize(dim, 1)
    return image, mask


def make_layer(image, mask, subsampler, options):
    image, mask = subsampler(image, mask, options)
    return djvu_support.photo_to_djvu(
        image=gamera_support.to_pil_rgb(image),
        mask_image=gamera_support.to_pil_1bpp(mask),
        slices=options.slices,
        crcb=options.crcb
    )


def image_dpi(image, options):
    dpi = options.dpi
    if dpi is None:
        dpi = image.dpi
    if dpi is None:
        dpi = djvu_support.DPI_DEFAULT
    dpi = max(dpi, djvu_support.DPI_MIN)
    dpi = min(dpi, djvu_support.DPI_MAX)
    return dpi


def image_to_djvu(width, height, image, mask, options):
    dpi = image_dpi(image, options)
    loss_level = options.loss_level
    if options.pages_per_dict > 1:
        # XXX This should probably go the other way round: we run minidjvu
        # first, and then reuse its masks.
        loss_level = 0
    sjbz = djvu_support.bitonal_to_djvu(gamera_support.to_pil_1bpp(mask), loss_level=loss_level)
    if options.fg_bg_defaults:
        image = gamera_support.to_pil_rgb(image)
        return djvu_support.Multichunk(width, height, dpi, image=image, sjbz=sjbz)
    else:
        chunks = dict(sjbz=sjbz)
        if options.foreground_options.slices != [0]:
            foreground_djvu = make_layer(image, mask, subsample_foreground, options.foreground_options)
            chunks.update(fg44=djvu_support.djvu_to_iw44(foreground_djvu))
        if options.background_options.slices != [0]:
            background_djvu = make_layer(image, mask, subsample_background, options.background_options)
            chunks.update(bg44=djvu_support.djvu_to_iw44(background_djvu))
        return djvu_support.Multichunk(width, height, dpi, **chunks)


def generate_mask(filename, image, method, kwargs):
    """
    Generate mask using the provided method (if filename is not None);
    or simply load it from file (if filename is not None).
    """
    if filename is None:
        return method(image, **kwargs)
    else:
        return gamera_support.load_image(filename)


def format_compression_info(bytes_in, bytes_out, bits_per_pixel):
    ratio = 1.0 * bytes_in / bytes_out
    percent_saved = (1.0 * bytes_in - bytes_out) * 100 / bytes_in
    msg = (
        f'{bits_per_pixel:.3f} bits/pixel; '
        f'{ratio:.3f}:1, {percent_saved:.2f}% saved, '
        f'{bytes_in} bytes in, {bytes_out} bytes out'
    )
    return msg


class Main:
    def __init__(self, prog=None):
        self._opened_files = []
        parser = cli.ArgumentParser(gamera_support.METHODS, default_method='djvu', prog=prog)
        parser.parse_arguments(actions=self)

    def __del__(self):
        for fd in self._opened_files:
            if not fd:
                continue
            try:
                fd.close()
            except Exception:
                pass

    def check_common(self, options):
        if len(options.masks) == 0:
            options.masks = [None for _ in options.input]
        elif len(options.masks) != len(options.input):
            error(
                'the number of input images ({0}) does not match the number of masks ({1})',
                len(options.input),
                len(options.masks)
            )
        # noinspection PyUnresolvedReferences
        log_level = {
            0: logging.WARNING,
            1: logging.INFO,
            2: logging.NOSY,
        }.get(options.verbosity, logging.DEBUG)
        LOGGER.setLevel(log_level)
        IPC_LOGGER.setLevel(log_level)
        djvu_support.require_cli()
        gamera_support.init()

    def check_multi_output(self, options):
        self.check_common(options)
        page_id_memo = {}
        if options.output is None:
            if options.output_template is not None:
                options.output = [
                    templates.expand(options.output_template, image_path, image_index, page_id_memo)
                    for image_index, image_path in enumerate(options.input)
                ]
                options.xmp_output = [f'{output_name}.xmp' if options.xmp else None for output_name in options.output]
            elif len(options.input) == 1:
                options.output = [sys.stdout]
                options.xmp_output = [None]
                check_tty()
            else:
                error('cannot output multiple files to stdout')
        else:
            if len(options.input) == 1:
                options.xmp_output = [f'{options.output}.xmp'] if options.xmp else [None]
                options.output = [options.output]
            else:
                error('cannot output multiple files to a single file')
        assert len(options.masks) == len(options.output) == len(options.input)
        options.output = [
            open(file_or_path, 'wb') if isinstance(file_or_path, str) else file_or_path
            for file_or_path in options.output
        ]
        options.xmp_output = [
            open(file_or_path, 'wb') if isinstance(file_or_path, str) else file_or_path
            for file_or_path in options.xmp_output
        ]
        self._opened_files.extend(options.output)
        self._opened_files.extend(options.xmp_output)

    def check_single_output(self, options):
        self.check_common(options)
        if options.output is None:
            options.output = [sys.stdout]
            options.xmp_output = [None]
            check_tty()
        else:
            filename = options.output
            # noinspection PyTypeChecker
            options.output = [open(filename, 'wb')]
            options.xmp_output = [open(f'{filename}.xmp', 'wb')] if options.xmp else [None]
            self._opened_files.extend(options.output)
            self._opened_files.extend(options.xmp_output)
        assert len(options.output) == len(options.xmp_output) == 1

    def encode(self, options):
        self.check_multi_output(options)
        parallel_for(options, self.encode_one, options.input, options.masks, options.output, options.xmp_output)

    def encode_one(self, options, image_filename, mask_filename, output, xmp_output):
        bytes_in = os.path.getsize(image_filename)
        LOGGER.info(f'{image_filename}:')
        file_type = filetype.check(image_filename)
        if file_type.like(filetype.Djvu):
            if file_type.like(filetype.DjvuSingle):
                LOGGER.info('- copying DjVu as is')
                with open(image_filename, 'rb') as djvu_file:
                    fs.copy_file(djvu_file, output)
            else:
                # TODO: Figure out how many pages the multi-page document
                #       consists of. If it's only one, continue.
                error('multi-page DjVu documents are not supported as input files')
            return
        LOGGER.info('- reading image')
        image = gamera_support.load_image(image_filename)
        width, height = image.ncols, image.nrows
        # noinspection PyUnresolvedReferences
        LOGGER.nosy(f'- image size: {width} x {height}')
        mask = generate_mask(mask_filename, image, options.method, options.parameters)
        n_connected_components = -1
        if xmp_output:
            n_connected_components = len(mask.cc_analysis())
        LOGGER.info('- converting to DjVu')
        djvu_doc = image_to_djvu(width, height, image, mask, options=options)
        djvu_file = djvu_doc.save()
        try:
            bytes_out = fs.copy_file(djvu_file, output)
        finally:
            djvu_file.close()
        bits_per_pixel = 8.0 * bytes_out / (width * height)
        compression_info = format_compression_info(bytes_in, bytes_out, bits_per_pixel)
        LOGGER.info(f'- {compression_info}')
        if xmp_output:
            LOGGER.info('- saving XMP metadata')
            metadata = xmp.metadata()
            metadata.import_(image_filename)
            internal_properties = list(cli.dump_options(options)) + [
                ('n-connected-components', str(n_connected_components))
            ]
            metadata.update(
                media_type='image/vnd.djvu',
                internal_properties=internal_properties,
            )
            metadata.write(xmp_output)
        djvu_doc.close()

    def separate_one(self, options, image_filename, output):
        LOGGER.info(f'{image_filename}:')
        file_type = filetype.check(image_filename)
        if file_type.like(filetype.Djvu):
            # TODO: Figure out how many pages the document consists of.
            #       If it's only one, extract the existing mask.
            error('DjVu documents are not supported as input files')
        LOGGER.info('- reading image')
        image = gamera_support.load_image(image_filename)
        width, height = image.ncols, image.nrows
        # noinspection PyUnresolvedReferences
        LOGGER.nosy(f'- image size: {width} x {height}')
        LOGGER.info('- thresholding')
        mask = generate_mask(None, image, options.method, options.parameters)
        LOGGER.info('- saving')
        if output is not sys.stdout:
            # A real file
            mask.save_PNG(output.name)
        else:
            tmp_output = temporary.file(suffix='.png')
            try:
                mask.save_PNG(tmp_output.name)
                fs.copy_file(tmp_output, output)
            finally:
                tmp_output.close()

    def separate(self, options):
        self.check_multi_output(options)
        for mask in options.masks:
            assert mask is None
        parallel_for(options, self.separate_one, options.input, options.output)

    def bundle(self, options):
        self.check_single_output(options)
        if (options.pages_per_dict <= 1) or (len(options.input) <= 1):
            self.bundle_simple(options)
        else:
            ipc.require('minidjvu')
            self.bundle_complex(options)
        [xmp_output] = options.xmp_output
        if xmp_output:
            LOGGER.info('saving XMP metadata')
            metadata = xmp.metadata()
            internal_properties = list(cli.dump_options(options, multi_page=True))
            metadata.update(
                media_type='image/vnd.djvu',
                internal_properties=internal_properties,
            )
            metadata.write(xmp_output)

    def _bundle_simple_page(self, o, input_file, mask, component_name):
        with open(component_name, 'wb') as component:
            self.encode_one(o, input_file, mask, component, None)

    def bundle_simple(self, options):
        [output] = options.output
        with temporary.directory() as tmpdir:
            bytes_in = 0
            component_filenames = []
            page_id_memo = {}
            for page, (input_file, mask) in enumerate(zip(options.input, options.masks)):
                bytes_in += os.path.getsize(input_file)
                page_id = templates.expand(options.page_id_template, input_file, page, page_id_memo)
                try:
                    djvu_support.validate_page_id(page_id)
                except ValueError as exception:
                    error(exception)
                component_filenames += [os.path.join(tmpdir, page_id)]
            parallel_for(options, self._bundle_simple_page, options.input, options.masks, component_filenames)
            LOGGER.info('bundling')
            djvu_file = djvu_support.bundle_djvu(*component_filenames)
            try:
                bytes_out = fs.copy_file(djvu_file, output)
            finally:
                djvu_file.close()
        bits_per_pixel = float('nan')  # FIXME!
        compression_info = format_compression_info(bytes_in, bytes_out, bits_per_pixel)
        # noinspection PyUnresolvedReferences
        LOGGER.nosy(compression_info)

    def _bundle_complex_page(self, options, page, minidjvu_in_dir, image_filename, mask_filename, pixels):
        LOGGER.info(f'{image_filename}:')
        file_type = filetype.check(image_filename)
        if file_type.like(filetype.Djvu):
            # TODO: Allow merging existing documents (even multi-page ones).
            error('DjVu documents are not supported as input files')
        LOGGER.info('- reading image')
        image = gamera_support.load_image(image_filename)
        dpi = image_dpi(image, options)
        width, height = image.ncols, image.nrows
        pixels[0] += width * height
        # noinspection PyUnresolvedReferences
        LOGGER.nosy(f'- image size: {width} x {height}')
        mask = generate_mask(mask_filename, image, options.method, options.parameters)
        LOGGER.info('- converting to DjVu')
        page.djvu = image_to_djvu(width, height, image, mask, options=options)
        del image, mask
        page.sjbz = djvu_support.Multichunk(width, height, dpi, sjbz=page.djvu['sjbz'])
        page.sjbz_symlink = os.path.join(minidjvu_in_dir, page.page_id)
        os.symlink(page.sjbz.save().name, page.sjbz_symlink)

    def bundle_complex(self, options):
        [output] = options.output
        with temporary.directory() as minidjvu_in_dir:
            bytes_in = 0
            pixels = [0]
            page_info = []
            page_id_memo: Dict[str, int] = {}
            for page_number, (image_filename, mask_filename) in enumerate(zip(options.input, options.masks)):
                page = utils.Namespace()
                page_info += [page]
                bytes_in += os.path.getsize(image_filename)
                page.page_id = templates.expand(options.page_id_template, image_filename, page_number, page_id_memo)
                try:
                    djvu_support.validate_page_id(page.page_id)  # type: ignore[attr-defined]
                except ValueError as exception:
                    error(exception)
            del page  # quieten pyflakes
            parallel_for(
                options,
                self._bundle_complex_page,
                page_info,
                itertools.repeat(minidjvu_in_dir),
                options.input,
                options.masks,
                itertools.repeat(pixels)
            )
            [pixel_count] = pixels
            with temporary.directory() as minidjvu_out_dir:
                LOGGER.info('creating shared dictionaries')

                def chdir() -> None:
                    os.chdir(minidjvu_out_dir)

                arguments = [
                    'minidjvu',
                    '--indirect',
                    '--pages-per-dict', str(options.pages_per_dict),
                ]
                if options.loss_level > 0:
                    arguments += ['--aggression', str(options.loss_level)]
                assert len(page_info) > 1  # minidjvu won't create single-page indirect documents
                arguments += [page.sjbz_symlink for page in page_info]  # type: ignore[attr-defined]
                index_filename = temporary.name(prefix='__index__.', suffix='.djvu', dir=minidjvu_out_dir)
                index_filename = os.path.basename(index_filename)  # FIXME: Name conflicts are still possible!
                arguments += [index_filename]
                ipc.Subprocess(arguments, preexec_fn=chdir).wait()
                os.remove(os.path.join(minidjvu_out_dir, index_filename))
                component_filenames = []
                for page_number, page in enumerate(page_info):
                    if page_number % options.pages_per_dict == 0:
                        iff_name = fs.replace_ext(page_info[page_number].page_id, 'iff')  # type: ignore[attr-defined]
                        iff_name = os.path.join(minidjvu_out_dir, iff_name)
                        component_filenames += [iff_name]
                    sjbz_name = os.path.join(minidjvu_out_dir, page.page_id)  # type: ignore[attr-defined]
                    component_filenames += [sjbz_name]
                    page.djvu['sjbz'] = sjbz_name  # type: ignore[attr-defined]
                    page.djvu['incl'] = iff_name  # type: ignore[attr-defined]
                    page.djvu = page.djvu.save()  # type: ignore[attr-defined]
                    page.djvu_symlink = os.path.join(minidjvu_out_dir, page.page_id)  # type: ignore[attr-defined]
                    os.unlink(page.djvu_symlink)  # type: ignore[attr-defined]
                    os.symlink(page.djvu.name, page.djvu_symlink)  # type: ignore[attr-defined]
                LOGGER.info('bundling')
                djvu_file = djvu_support.bundle_djvu(*component_filenames)
                try:
                    bytes_out = fs.copy_file(djvu_file, output)
                finally:
                    djvu_file.close()
        bits_per_pixel = 8.0 * bytes_out / pixel_count
        compression_info = format_compression_info(bytes_in, bytes_out, bits_per_pixel)
        # noinspection PyUnresolvedReferences
        LOGGER.nosy(compression_info)


__all__ = ['Main']



# Copyright © 2010-2018 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

from setuptools import setup, find_packages
from pathlib import Path


ROOT_DIRECTORY = Path(__file__).parent.resolve()


def get_version():
    changelog = ROOT_DIRECTORY / 'doc' / 'changelog'
    with open(changelog, mode='r') as fd:
        line = fd.readline()
    return line.split()[1].strip('()')


setup(
    name='didjvu',
    description='DjVu encoder with foreground/background separation (Python 3 fork) ',
    version=get_version(),
    license='GPL-2.0-only',
    long_description=(ROOT_DIRECTORY / 'README.rst').read_text(encoding='utf-8'),
    long_description_content_type='text/x-rst',
    author='Jakub Wilk',
    maintainer='FriedrichFröbel',
    url='https://github.com/FriedrichFroebel/didjvu/',
    packages=find_packages(
        where='.',
        exclude=['tests', 'tests.*', 'private', 'private.*']
    ),
    include_package_data=True,
    python_requires='>=3.10, <4',
    install_requires=[
        'gamera>=4.0.0',
        'Pillow',
    ],
    extras_require={
        'dev': [
            'coverage',
            'flake8',
            'pep8-naming',
        ],
        'docs': [
            'docutils',
            'pygments',
        ]
    },
    entry_points={
        'console_scripts': [
            'didjvu=didjvu.__main__:main',
        ],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Text Processing',
        'Topic :: Multimedia :: Graphics',
    ]
)



# Copyright © 2015-2018 Jakub Wilk <jwilk@jwilk.net>
# Copyright © 2022-2024 FriedrichFroebel
#
# This file is part of didjvu.
#
# didjvu is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# didjvu is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.

import os
import re
from xml.etree import ElementTree

from tests.tools import TestCase

from didjvu import version

_HERE = os.path.dirname(__file__)
DOC_DIRECTORY = os.path.join(_HERE, os.pardir, 'doc')


class ChangelogTestCase(TestCase):
    def test_changelog(self):
        path = os.path.join(DOC_DIRECTORY, 'changelog')
        with open(path, 'rt') as file:
            line = file.readline()
        changelog_version = line.split()[1].strip('()')
        self.assertEqual(changelog_version, version.__version__)


class ManpageTestCase(TestCase):
    def test_manpage(self):
        path = os.path.join(DOC_DIRECTORY, 'manpage.xml')
        # Do not feed path into `iterparse()` due to
        # https://github.com/python/cpython/issues/69893
        with open(path, mode='rb') as fd:
            for event, element in ElementTree.iterparse(fd):
                if element.tag == 'refmiscinfo' and element.get('class') == 'version':
                    self.assertEqual(element.text, version.__version__)
                    break
            else:
                self.fail("missing <refmiscinfo class='version'>")


class GetSoftwareAgentTestCase(TestCase):
    def test_get_software_agent(self):
        result = version.get_software_agent()
        self.assertIsInstance(result, str)
        match = re.match(r'^didjvu [\d.]+ [(]Gamera [\d.]+(pre)?[)]$', result)
        self.assertIsNotNone(match)
